//==========================================================================
//
// au1500_gpio.c
//
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         02/05/2002
// Description:  This file contains code to intialize the Au1500 GPIO
//               section as well as set, clear and test functions
//               for manipulating the GPIO bits
//
//--------------------------------------------------------------------------

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "bits.h"
#include "au1500_gpio.h"

// Function Prototypes
void GPIO_init(void);
int GPIO_set(int);
int GPIO_clr(int);
int GPIO_tst(int);
int GPIO_in(int);
int GPIO_out(int);

//--------------------------------------------------------
// GPIO_init()
//
// This function sets the startup state for the Au1500
// GPIO Ports 1 and 2
//
void
GPIO_init()
{

	// GPIO1 init
	// for GPIO1 we use most of the bits for their peripheral function
	// not as GPIO.  The exceptions are GPPIO0, 1 and 4 which we use for
	// the user pushbutton switches, and GPIO5-8 which we use as interrupts
	// for various peripherals.  Although umon does not use interrupts, we
	// still enable them as inputs here.

	// set the default alternate functions first
	GPIO1(GPIO1_PINFUNC) = GPIO1_PF_USB			// USB Host
						 | GPIO1_PF_EX1			// EXTCLK1
						 | GPIO1_PF_EX0			// EXTCLK0
						 | GPIO1_PF_U3_CTL;		// UART3 Control

	GPIO1(GPIO1_TRI) = 0xffffffff;				// tristate all of GPIO 1

	GPIO1(GPIO1_INPUTEN) = 0;					// enable GPIO1 inputs

	// GPIO2 init - enable the GPIO2 block
	GPIO2(GPIO2_ENABLE)	= GPIO2_EN_CE;

	// Set the default outputs and directions
	GPIO2(GPIO2_DIR) = 0;	// start GPIO2 as all inputs

	// set all of the low true outputs (except PRST) to a high state before we enable them
	// remember we need to set both the enable bits and the output bit to actually set an output
	GPIO2(GPIO2_OUTPUT) = 0xffff0000
						| GPIO_NUM2BIT(GPIO_USBH_PWR_EN)
						| GPIO_NUM2BIT(GPIO_USER_LED0)	
						| GPIO_NUM2BIT(GPIO_USER_LED1)	
						| GPIO_NUM2BIT(GPIO_USER_LED2)	;

	// now enable all of the outputs
	GPIO2(GPIO2_DIR) = GPIO_NUM2BIT(GPIO_PCI_RST)
					 | GPIO_NUM2BIT(GPIO_USBH_PWR_EN)
					 | GPIO_NUM2BIT(GPIO_USER_LED0)
					 | GPIO_NUM2BIT(GPIO_USER_LED1)
					 | GPIO_NUM2BIT(GPIO_USER_LED2)
					 | GPIO_NUM2BIT(GPIO_PCI_ID0)
					 | GPIO_NUM2BIT(GPIO_PCI_ID1);
}

//--------------------------------------------------------
// GPIO_set()
//
// This function sets the desired bit passed in.  It first 
// tests to see if the bit is an actual output.  If not,
// it returns a -1, otherwise it returns 0.
//

int GPIO_set(int gpio_bit)
{
	// quick sanity test
	if ((gpio_bit > 215) || (gpio_bit < 200)) return -1;

	gpio_bit = gpio_bit - 200;

	// Get the direction and see if it's an output
	if (!(GPIO2(GPIO2_DIR) & (1 << gpio_bit))){
		printf("Attempted to set GPIO %d, while configured as input!\n", (200 + gpio_bit));
		return -1;	// can't set it if it's not an output!
	}									 
	// if yes, then set it
	GPIO2(GPIO2_OUTPUT) |= (1 << gpio_bit);

	return 0;
}

//--------------------------------------------------------
// GPIO_clr()
//
// This function clears the desired bit passed in.  It first 
// tests to see if the bit is an actual output.  If not,
// it returns a -1, otherwise it returns 0.  NOTE: We do not
// test to see if clearing the bit would screw up any alternate
// functions.  Use this function with caution!
//

int GPIO_clr(int gpio_bit)
{
	// quick sanity test
	if ((gpio_bit > 215) || (gpio_bit < 200)) return -1;

	gpio_bit = gpio_bit - 200;

	// Get the direction and see if it's an output
	if (!(GPIO2(GPIO2_DIR) & (1 << gpio_bit))){
		printf("Attempted to set GPIO %d, while configured as input!\n", (200 + gpio_bit));
		return -1;	// can't set it if it's not an output!
	}

	// if yes, then clear it
	GPIO2(GPIO2_OUTPUT) &= ~(1 << gpio_bit);

	return 0;
}
//--------------------------------------------------------
// GPIO_tst()
//
// This function returns the state of desired bit passed in.
// It does not test to see if it's an input or output and thus
// can be used to verify if an output set/clr has taken place
// as well as for testing an input state.
//

int GPIO_tst(int gpio_bit)
{
	// quick sanity test
	if ((gpio_bit > 215) || (gpio_bit < 200)) return -1;

	gpio_bit = gpio_bit - 200;

	// get the state
	if (GPIO2(GPIO2_PINSTATE) & (1 << gpio_bit)) return 1;

	return 0;

}

//--------------------------------------------------------
// GPIO_in()
//
// This function changes the direction of the desired bit 
// to input.  NOTE: We do not test to see if changing the
// direction of the bit would screw up anything.  Use this
// function with caution!
int GPIO_in(int gpio_bit)
{
	// quick sanity test
	if ((gpio_bit > 215) || (gpio_bit < 200)) return -1;

	gpio_bit = gpio_bit - 200;

	// clear the direction bit
	GPIO2(GPIO2_DIR) &= ~(1 << gpio_bit);

	return 0;
}

//--------------------------------------------------------
// GPIO_out()
//
// This function changes the direction of the desired bit 
// to output.  NOTE: We do not test to see if changing the
// direction of the bit would screw up anything.  Use this
// function with caution!
int GPIO_out(int gpio_bit)
{
	// quick sanity test
	if ((gpio_bit > 215) || (gpio_bit < 200)) return -1;

	gpio_bit = gpio_bit - 200;

	// set the direction bit
	GPIO2(GPIO2_DIR) |= (1 << gpio_bit);

	return 0;
}

