//==========================================================================
//
//      cpu_pci.c
//
//      CPU Specific PCI support and initialization routines
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         02/05/2002
// Description:  This file contains PCI intialization and support routines 
//				 specific to the 405GP
//
//--------------------------------------------------------------------------

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "pci.h"
#include "cpu_pci.h"

#define PCI_DBG

//--------------------------------------------------------------------------
// Function Prototypes
//
void set_cfg_type_id(ulong bus, ulong dev_func);
ulong pci_cfg_read_dword (ulong bus, ulong dev_func, ulong offset);
ushort pci_cfg_read_word (ulong bus, ulong dev_func, ulong offset);
uchar pci_cfg_read_byte (ulong bus, ulong dev_func, ulong offset);
void pci_cfg_write_dword (ulong bus, ulong dev_func, ulong offset, ulong data);
void pci_cfg_write_word (ulong bus, ulong dev_func, ulong offset, ushort data);
void pci_cfg_write_byte (ulong bus, ulong dev_func, ulong offset, uchar  data);
void pci_init(void);

extern void udelay(int);
extern int pci_probe_bus(void);

// shared memory base address for all PCI devices
ulong sh_mem_base;

//--------------------------------------------------------------------------
// PCI Support routines - specific to 405GP Bridge
//

//--------------------------------------------------------------------------
// Configuration Space read/write routines
// 
ulong pci_cfg_read_dword (ulong bus, ulong dev_func, ulong offset)
{
    ulong config_dword;

	// select the device and make sure we have the right config type
	set_cfg_type_id(bus, dev_func);

#ifdef PCI_DBG
	printf("pci_cfg_read_dword() Bus: %ld, Dev: %ld, Func: %ld, ID: %02x\n", 
		 bus, PCI_GET_DEV(dev_func), PCI_GET_FUNC(dev_func), (char)IDSEL);
#endif
	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN 		
					| PCI_CFG_BUS(bus)	
					| PCI_CFG_DEV(PCI_GET_DEV(dev_func)) 
					| PCI_CFG_FUNC(PCI_GET_FUNC(dev_func))
					| PCI_CFG_REG(offset));

	config_dword = HTOLE32(PCI_CFG_DATA_32(offset));
	asm volatile ("eieio");

#ifdef PCI_DBG
	printf("pci_cfg_read_dword() CFG_ADD: %08lx, CFD_DATA: %08lx\n", PCI_CFG_ADD, config_dword);
#endif

//    if (pci_master_abort()) return 0xffffffff;

    return config_dword;
}

ushort pci_cfg_read_word (ulong bus, ulong dev_func, ulong offset)
{
    ushort config_word;

	set_cfg_type_id(bus, dev_func);

#ifdef PCI_DBG
	printf("pci_cfg_read_word() Bus: %ld, Dev: %ld, Func: %ld, ID: %02x\n", 
		bus, PCI_GET_DEV(dev_func), PCI_GET_FUNC(dev_func), (char)IDSEL);
#endif
	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN 		
					| PCI_CFG_BUS(bus)	
					| PCI_CFG_DEV(PCI_GET_DEV(dev_func)) 
					| PCI_CFG_FUNC(PCI_GET_FUNC(dev_func))
					| PCI_CFG_REG(offset));

	config_word = HTOLE16(PCI_CFG_DATA_16(offset));
	asm volatile ("eieio");

#ifdef PCI_DBG
	printf("pci_cfg_read_word() CFG_ADD: %08lx, CFD_DATA: %04x\n", PCI_CFG_ADD, config_word);
#endif
//    if (pci_master_abort()) return 0xffff;

    return config_word;
}

uchar pci_cfg_read_byte (ulong bus, ulong dev_func, ulong offset)
{
    uchar config_byte;

	set_cfg_type_id(bus, dev_func);

#ifdef PCI_DBG
	printf("pci_cfg_read_byte() Bus: %ld, Dev: %ld, Func: %ld, ID: %02x\n", 
		bus, PCI_GET_DEV(dev_func), PCI_GET_FUNC(dev_func), (char)IDSEL);
#endif
	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN 		
					| PCI_CFG_BUS(bus)	
					| PCI_CFG_DEV(PCI_GET_DEV(dev_func)) 
					| PCI_CFG_FUNC(PCI_GET_FUNC(dev_func))
					| PCI_CFG_REG(offset));

	config_byte = PCI_CFG_DATA_8(offset);
	asm volatile ("eieio");

#ifdef PCI_DBG
	printf("pci_cfg_read_byte() CFG_ADD: %08lx, CFD_DATA: %02x\n", PCI_CFG_ADD, config_byte);
#endif
//    if (pci_master_abort()) return 0xff;

    return config_byte;
}

void pci_cfg_write_dword (ulong bus, ulong dev_func, ulong offset, ulong data)
{

	set_cfg_type_id(bus, dev_func);

#ifdef PCI_DBG
	printf("pci_cfg_write_dword() Bus: %ld, Dev: %ld, Func: %ld, ID: %02x\n", 
		bus, PCI_GET_DEV(dev_func), PCI_GET_FUNC(dev_func), (char)IDSEL);
#endif
	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN 		
					| PCI_CFG_BUS(bus)	
					| PCI_CFG_DEV(PCI_GET_DEV(dev_func)) 
					| PCI_CFG_FUNC(PCI_GET_FUNC(dev_func))
					| PCI_CFG_REG(offset));

    PCI_CFG_DATA_32(offset) =  HTOLE32(data);
	asm volatile ("eieio");

#ifdef PCI_DBG
	printf("pci_cfg_write_dword() CFG_ADD: %08lx, CFD_DATA: %08lx\n", PCI_CFG_ADD, data);
#endif
}

void pci_cfg_write_word (ulong bus, ulong dev_func, ulong offset, ushort data)
{
	set_cfg_type_id(bus, dev_func);

#ifdef PCI_DBG
	printf("pci_cfg_write_word() Bus: %ld, Dev: %ld, Func: %ld, ID: %02x\n", 
		bus, PCI_GET_DEV(dev_func), PCI_GET_FUNC(dev_func), (char)IDSEL);
#endif
	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN 		
					| PCI_CFG_BUS(bus)	
					| PCI_CFG_DEV(PCI_GET_DEV(dev_func)) 
					| PCI_CFG_FUNC(PCI_GET_FUNC(dev_func))
					| PCI_CFG_REG(offset));

    PCI_CFG_DATA_16(offset) =  HTOLE16(data);
	asm volatile ("eieio");

#ifdef PCI_DBG
	printf("pci_cfg_write_word() CFG_ADD: %08lx, CFD_DATA: %04x\n", PCI_CFG_ADD, data);
#endif
}

void pci_cfg_write_byte (ulong bus, ulong dev_func, ulong offset, uchar  data)
{
	set_cfg_type_id(bus, dev_func);

#ifdef PCI_DBG
	printf("pci_cfg_write_byte() Bus: %ld, Dev: %ld, Func: %ld, ID: %02x\n", 
		bus, PCI_GET_DEV(dev_func), PCI_GET_FUNC(dev_func), (char)IDSEL);
#endif
	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN 		
					| PCI_CFG_BUS(bus)	
					| PCI_CFG_DEV(PCI_GET_DEV(dev_func)) 
					| PCI_CFG_FUNC(PCI_GET_FUNC(dev_func))
					| PCI_CFG_REG(offset));

    PCI_CFG_DATA_8(offset) =  data;
	asm volatile ("eieio");

#ifdef PCI_DBG
	printf("pci_cfg_write_byte() CFG_ADD: %08lx, CFD_DATA: %02x\n", PCI_CFG_ADD, data);
#endif
}

//--------------------------------------------------------------------------
// set_cfg_type_id()
//
// This support function sets the ID bit and the correct config type
// It is used by the various pci_cfg_read_xx and write_xx
// functions that follow.
void set_cfg_type_id(ulong bus, ulong dev_func){

	GPIO0_OR = GPIO0_OR & ~(IDSEL1 | IDSEL2);	// clear idsel bits
	// if bus is non-zero, it means we have to do a type 1 config cycle
	// to reach through the device (which must be a bridge) to the
	// other bus.  The actual bus number does not matter since any non-zero
	// bus by definition is on the other side of a bridge.
	if (bus != 0) {
	}
	else {
		// Bus 0
		// convert a device number into a single bit in the ID Select register
		// device 0 is a special case that selects the 405 registers
		if(PCI_GET_DEV(dev_func)){
			switch(PCI_GET_DEV(dev_func))
			{
				case DEV_START + 0 : GPIO0_OR |= IDSEL1; break;
				case DEV_START + 1 : GPIO0_OR |= IDSEL2; break;
				default : break;
			}
		}
	}
}

//-------------------------------------------------------------------------
// Initialize the bridge
//
void pci_init(void)
{
	printf("Initializing the 405GP/r PCI Bridge\n");

	// enable the tristate for the idsel BITS gpio 15 AND 16
	GPIO0_TS |= 0x00018000;

	// Setup the PLB to PCI side memory
	// window

	// 1:1 mapping local to pci address
	PCIL0_PMM0PCILA = HTOLE32(PCI_MEM_BASE);

	// 32-bit addressing only
	PCIL0_PMM0PCIHA = 0x0;

	PCIL0_PMM0LA = HTOLE32(PCI_MEM_BASE);

	// set the mask bits and enable the window - no prefetching
	PCIL0_PMM0MA = HTOLE32((~(PCI_MEM_SIZE - 1)) | PCIL0_PMM0MA_ENA);

	// I/O window is fixed at 0xE8800000
	// so no setup is available

	// now setup the PCI to Local side memory
	// window, 1:1, starting at 0, 256Mbytes.
	// PCI side window address is setup using
	// config cycles, while local side remap
	// address is directly addressed.
	// PCI config cycles to Bus 0 Dev 0 will
	// select 405GP/r.

	// Enable BAR and set PCI Side window address to 0.
	pci_cfg_write_dword (0, 0, PCIC0_PTM1BAR, 0x01);

	// enable 405 for memory accesses and mastering
	pci_cfg_write_word (0, 0, PCIC0_CMD, (PCI_CFG_CMD_MEM | PCI_CFG_CMD_MASTER));

	// from local side, set BAR0 size to 256Mbytes, enable it
	// and set remap to 0
	PCIL0_PTM1LA = 0x0;
	PCIL0_PTM1MS = HTOLE32(0xF0000000 | PCIL0_PTM1MS_ENA);
 
	// reset the PCI bus
//	temp16 = pci_cfg_read_word (0, 0, PCIC0_BRDGOPT2);
//	pci_cfg_write_word (0, 0, PCIC0_BRDGOPT2, temp16 | PCIC0_BRDGOPT2_DPR);
//	udelay(1000);

//	while(1){}
	// release reset
//	pci_cfg_write_word (0, 0, PCIC0_BRDGOPT2, temp16);

	printf("PCI Initialization Complete\n");

	// initialize the shared memory base address
	sh_mem_base = PCI_SM_BASE;

}

//--------------------------------------------------------------------------
// End of cpu_pci.c
