//=============================================================================
//
//      cpu_pci.h
//
//      PCI Support
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         2002-05-21
// Description:  This file contains the PCI defines and access macros 
//				 for the MPC8245/CSB281.  It also contains the defines for
//				 the MPC8245 configuration registers that are accessed
//				 via PCI config cycles
//
//
//=============================================================================

#include "bits.h"

// #define PCI_DBG

//-----------------------------------------------------------------------------
// 8245 will not pass through config cycles for device numbers less than 11   
#define DEV_START	11

//-----------------------------------------------------------------------------
// Address of PCI Configuration Address and Data Registers
#define PCI_CFG_ADD_REG			0xfec00000
#define PCI_CFG_DATA_REG		0xfee00000
#define PCI_CFG_EN 				BIT31	// Enable BIT
#define PCI_CFG_ADD  			*(vulong *) (PCI_CFG_ADD_REG)
#define PCI_CFG_DATA_32(_x_)  	*(vulong *) (PCI_CFG_DATA_REG | (_x_ & 0x3))
#define PCI_CFG_DATA_16(_x_)  	*(vushort *)(PCI_CFG_DATA_REG |(_x_ & 0x3))
#define PCI_CFG_DATA_8(_x_)  	*(vuchar *) (PCI_CFG_DATA_REG | (_x_ & 0x3))

//-----------------------------------------------------------------------------
// Mappings for PCI memory and IO spaces
//

// The MPC8245 implements 1 Local bus to PCI Aperture for Memory Space, an
// aperture for Local to PCI I/O Space, 0 to 64KB PCI I/O Addresses and another
// apeture for Local to PCI I/O Space, 1MB to 8MB PCI I/O Addresses.
// Configuration is handled via a Config Address Register and a Config Data
// Register. IDSEL is handled via the external BCSR (Board Control/Status Reg.)
//
// These are the addresses, in CPU space, where the PCI apertures are
// for each type of PCI access.  This is also where the PCI devices
// will be mapped when found.
#define PCI_MEM_BASE 		0x80000000		// fixed by MPC8245 bridge
#define PCI_MEM_SIZE		0x40000000		// 1G size

#define PCI_IO_BASE 		0xFE800000		// fixed by MPC8245 bridge
#define PCI_IO_SIZE			0x00400000		// 4Mbyte size

#define PCI_SM_BASE			0x00100000		// 2nd 1Meg of SDRAM is shared memory space
#define PCI_SM_SIZE			0x00100000		// 1Mbyte size

//--------------------------------------------------------------------------
// Host to Little Endian Macros
// Big Endian Host
#define HTOLE16(x) ((((x) & 0xff00U) >> 8) | (((x) & 0x00ffU) << 8))
#define HTOLE32(x) ((((x) & 0xff000000UL) >> 24) | (((x) & 0x00ff0000UL) >> 8 ) | (((x) & 0x0000ff00UL) << 8 ) | (((x) & 0x000000ffUL) << 24))

// Little Endian Host or Big Endian Host that Swaps on it's own (like the 405GP)
//# define HTOLE16(x) (x)
//# define HTOLE32(x) (x)

//-----------------------------------------------------------------------------
// Macros for reading and writing the shared memory structures
//-----------------------------------------------------------------------------
#define RD_SM8(_reg_,_val_)  ((_val_) = *((vuchar *)((((ulong)_reg_) ^ (0)))))
#define WR_SM8(_reg_,_val_)  (*((vuchar *)((((ulong)_reg_) ^ (0)))) = (_val_))

#define RD_SM16(_reg_,_val_) ((_val_) = HTOLE16(*((vushort *)((((ulong)_reg_) ^ (0))))))
#define WR_SM16(_reg_,_val_) (*((vushort *)((((ulong)_reg_) ^ (0)))) = HTOLE16(_val_))

#define RD_SM32(_reg_,_val_) ((_val_) = HTOLE32(*((vulong *)((ulong)_reg_))))
#define WR_SM32(_reg_,_val_) (*((vulong *)((ulong)_reg_)) = HTOLE32(_val_))

//-----------------------------------------------------------------------------
// MPC8245 Configuration Registers accessed from PCI Space.
// Use config bus 0 and device 0 to access.	 
//
// Common PCI Registers	- Use the standard PCI names defined in
// pci.h instead of these.  They are shown here FYI.
#define PCI8245_VENDID 				0x00 		// PCI Vendor ID
#define PCI8245_DEVID 				0x02 		// PCI Device ID
#define PCI8245_CMD 				0x04 		// PCI Command Register
#define PCI8245_STATUS 				0x06 		// PCI Status Register
#define PCI8245_REVID 				0x08 		// PCI Revision ID
#define PCI8245_PCICLS				0x09 		// PCI Class Register
#define PCI8245_CACHELS 			0x0C 		// PCI Cache Line Size
#define PCI8245_LATTIM				0x0D 		// PCI Latency Timer
#define PCI8245_HDTYPE 				0x0E 		// PCI Header Type
#define PCI8245_BIST 				0x0F 		// PCI Built In Self Test Control
#define PCI8245_PTM1BAR 			0x14 		// PCI PTM 1 BAR
#define PCI8245_PTM2BAR 			0x18 		// PCI PTM 2 BAR
#define PCI8245_SBSYSVID 			0x2C 		// PCI Subsystem Vendor ID
#define PCI8245_SBSYSID 			0x2E 		// PCI Subsystem ID
#define PCI8245_CAP 				0x34 		// PCI Capabilities Pointer
#define PCI8245_INTLN 				0x3C 		// PCI Interrupt Line
#define PCI8245_INTPN 				0x3D 		// PCI Interrupt Pin
#define PCI8245_MINGNT 				0x3E 		// PCI Minimum Grant
#define PCI8245_MAXLTNCY 			0x3F 		// PCI Maximum Latency
#define PCI8245_PCI_GCR 			0x44 		// PCI General Control
// end of PCI common registers
//
// Extended PCI Registers - MPC8245 specific
#define PCI8245_PCI_ARB				0x46		// PCI arbiter control register
#define PCI8245_PMCR1				0x70 		// Power management configuration register 1
#define PCI8245_PMCR2				0x72 		// Power management configuration register 2
#define PCI8245_ODCR				0x73 		// Output driver control register 1
#define PCI8245_CDCR				0x74 		// CLK driver control register 2
#define PCI8245_MDCR1				0x76 		// Miscellaneous driver control register 1
#define PCI8245_MDCR2				0x77 		// Miscellaneous driver control register 2
#define PCI8245_EUMB_BASE			0x78 		// Embedded utilities memory block base address register
#define PCI8245_MEM_START1			0x80 		// Memory starting address registers
#define PCI8245_MEM_START2			0x84
#define PCI8245_XMEM_START1			0x88 		// Extended memory starting address registers
#define PCI8245_XMEM_START2			0x8C
#define PCI8245_MEM_END1			0x90 		// Memory ending address registers
#define PCI8245_MEM_END2			0x94
#define PCI8245_XMEM_END1			0x98 		// Extended memory ending address registers
#define PCI8245_XMEM_END2			0x9C
#define PCI8245_MEM_EN				0xA0		// Memory bank enable register
#define PCI8245_PGMAX				0xA3		// Page mode counter/timer
#define PCI8245_PIF_CFG1			0xA8		// Processor interface configuration 1
#define PCI8245_PIF_CFG2			0xAC		// Processor interface configuration 2
#define PCI8245_ECC_ERR_CTR			0xB8		// ECC single bit error counter
#define PCI8245_ECC_ERR_TRG			0xB9		// ECC single bit error trigger register
#define PCI8245_ERR_EN1				0xC0		// Error enabling register 1
#define PCI8245_ERR_DET1			0xC1		// Error detection register 1
#define PCI8245_PBUS_ERR			0xC3		// Processor internal bus error status
#define PCI8245_ERR_EN2				0xC4		// Error enabling register 2
#define PCI8245_ERR_DET2			0xC5		// Error detection register 2
#define PCI8245_PCI_ERR				0xC7		// PCI bus error status register
#define PCI8245_ERR_ADD				0xC8		// Processor/PCI error address register
#define PCI8245_XROM_CFG1			0xD0		// Extended ROM configuration register
#define PCI8245_XROM_CFG2			0xD4		// Extended ROM configuration register 2
#define PCI8245_XROM_CFG3			0xD8		// Extended ROM configuration register 3
#define PCI8245_XROM_CFG4			0xDC		// Extended ROM configuration register 4
#define PCI8245_MAPB_OPT			0xE0		// Address map B options register
#define PCI8245_PLL_CFG				0xE2		// PLL configuration register
#define PCI8245_MCCR1 				0xF0		// Memory Control/Configuration register 1
#define PCI8245_MCCR2 				0xF4		// Memory Control/Configuration register 2
#define PCI8245_MCCR3 				0xF8		// Memory Control/Configuration register 3
#define PCI8245_MCCR4 				0xFC		// Memory Control/Configuration register 4

// Bit defines for Extended PCI Registers
//
// PCI8245_PCI_ARB - PCI arbiter control register
#define PCI_ARB_ON_CHIP			BIT15	// 1 = Enable on-chip PCI arbitration.
#define PCI_ARB_PARK_LAST		0x00 << 13	// Parking mode when there are no bus requests and bus is idle.
#define PCI_ARB_PARK_DEV0		0x01 << 13	// 0x00 = Parked with last device, 0x01 = Parked with device on REQ0/GNT0.
#define PCI_ARB_PARK_8245		0x02 << 13	// 0x10 = Parked with MPC8245.
#define PCI_ARB_GRANT_WAIT		BIT12		// 0 = Arbiter negates GNTx signal to master if no FRAME within 16 clocks,
											// 1 = wait until FRAME asserted or REQ goes away
#define PCI_ARB_CFG_RETRY		BIT10 		// 1 = PCI target logic retries all external PCI configuration transactions.
											// 0 = PCI target logic responds to external PCI configuration transactions.
#define PCI_ARB_PRI8245			BIT7 		// MPC8245 priority level, 1 = high, 0 = low
#define PCI_ARB_PRI_DEV4		BIT4		// External device priority levels, 1 = high, 0 = low
#define PCI_ARB_PRI_DEV3		BIT3		// External device priority levels, 1 = high, 0 = low
#define PCI_ARB_PRI_DEV2		BIT2		// External device priority levels, 1 = high, 0 = low
#define PCI_ARB_PRI_DEV1		BIT1		// External device priority levels, 1 = high, 0 = low
#define PCI_ARB_PRI_DEV0		BIT0		// External device priority levels, 1 = high, 0 = low

// PCI8245_PMCR1 - Power management configuration register 1
#define PMCR1_NAP_MSG 			BIT15		// SW must set bit to indicate MPC8245 doesn't broadcast HALT message prior to nap mode
#define PMCR1_SLEEP_MSG 		BIT14		// SW must set bit to indicate MPC8245 doesn't broadcast SLEEP message prior to sleep mode
#define PMCR1_LP_REF_EN 		BIT12		// 1 = MPC8245 performs memory refresh cycles when in sleep mode
#define PMCR1_PMEN 				BIT7		// 1 = Enables peripheral logic power management
#define PMCR1_BR1_WAKE			BIT6		// 1 = BR1 causes the MPC8245 to wake up from nap or sleep mode
#define PMCR1_DOZE 				BIT5		// 1 = Enables the doze mode (if PMCR1_PMEN is set)
#define PMCR1_NAP 				BIT4		// 1 = Enables the nap mode (if PMCR1_PMEN is set)
#define PMCR1_SLEEP 			BIT3		// 1 = Enables the sleep mode (if PMCR1_PMEN is set)
#define PMCR1_CKO_OFF			0x00 << 1 	// Clock source for test clock output when CKO_SEL = 1, 0x00 = off
#define PMCR1_CKO_SYSCLK		0x01 << 1 	//   0x01 Selects the internal sys_logic_clk signal
#define PMCR1_CKO_PCICLK_2		0x02 << 1 	//   0x02 Selects one-half of the PCI rate clock
#define PMCR1_CKO_PCICLK		0x03 << 1 	//   0x03 Selects the internal PCI rate clock
#define PMCR1_CKO_SEL			BIT0		// 0 = CKO = Processor core clock, 1 = System logic clock selected CKO_MODE bits above.

// PCI8245_PMCR2 - Power management configuration register 2
#define PMCR2_DLL_EXTEND 		BIT7		// 1 = DLL extended range
#define PMCR2_PCI_HOLD_66		0x00 << 4	// PCI output hold delay value, 00 Recommended for 66 MHz PCI bus (default)
#define PMCR2_PCI_HOLD_33		0x02 << 4	//   10 Recommended for 33 MHz PCI bus
#define PMCR2_PCI_PLL_SLP 		BIT2		// 1 = The MPC8245 samples the PLL configuration pins coming out of sleep mode

// PCI8245_ODCR - Output driver control register 1
#define ODCR_DRV_PCI_40			0x00 << 7	// 40 ohm drive PCI/EPIC signals
#define ODCR_DRV_PCI_20			0x01 << 7	// 20 ohm on all except 6 ohm on FRAME and S_RST
#define ODCR_DRV_MEM_STD_40		0x00 << 6	// 40 ohm drive on standard and memory non-control signals
#define ODCR_DRV_MEM_STD_20		0x01 << 6	// 20 ohm drive on standard and memory non-control signals
#define ODCR_DRV_MEM_CTL_40		0x01 << 4	// 40 ohm drive on memory control signals
#define ODCR_DRV_MEM_CTL_20		0x02 << 4	// 20 ohm drive on memory control signals
#define ODCR_DRV_MEM_CTL_6 		0x03 << 4	// 6 ohm drive on memory control signals
#define ODCR_DRV_PCI_CLK_40		0x01 << 2	// 40 ohm drive on PCI_CLK[0:4] and PCI_CLK_SYNC_OUT.
#define ODCR_DRV_PCI_CLK_20		0x02 << 2	// 20 ohm drive on PCI_CLK[0:4] and PCI_CLK_SYNC_OUT.
#define ODCR_DRV_PCI_CLK_6		0x03 << 2	// 6 ohm drive on PCI_CLK[0:4] and PCI_CLK_SYNC_OUT.
#define ODCR_DRV_MEM_CLK_40		0x01 << 0	// 40 ohm drive on SDRAM_CLK[0:4] and SDRAM_CLK_SYNC_OUT.
#define ODCR_DRV_MEM_CLK_20		0x02 << 0	// 20 ohm drive on SDRAM_CLK[0:4] and SDRAM_CLK_SYNC_OUT.
#define ODCR_DRV_MEM_CLK_6		0x03 << 0	// 6 ohm drive on SDRAM_CLK[0:4] and SDRAM_CLK_SYNC_OUT.

// PCI8245_CDCR - CLK driver control register 2
#define CDCR_PCI_SYNC_OUT_DIS	BIT15		// 1 = Disable PCI_SYNC_OUT
#define CDCR_PCI_CLK0_DIS		BIT14		// 1 = Disable PCI_CLK0
#define CDCR_PCI_CLK1_DIS		BIT13		// 1 = Disable PCI_CLK1
#define CDCR_PCI_CLK2_DIS		BIT12		// 1 = Disable PCI_CLK2
#define CDCR_PCI_CLK3_DIS		BIT11		// 1 = Disable PCI_CLK3
#define CDCR_PCI_CLK4_DIS		BIT10		// 1 = Disable PCI_CLK4
#define CDCR_SDRAM_CLK0_DIS		BIT6 		// 1 = Disable SDRAM_CLK0
#define CDCR_SDRAM_CLK1_DIS		BIT5 		// 1 = Disable SDRAM_CLK1
#define CDCR_SDRAM_CLK2_DIS		BIT4 		// 1 = Disable SDRAM_CLK2
#define CDCR_SDRAM_CLK3_DIS		BIT3 		// 1 = Disable SDRAM_CLK3

// PCI8245_MDCR1 - Miscellaneous driver control register 1
#define MDCR1_MCP_OD 			BIT7		// 1 = MCP output is open-drain.
#define MDCR1_SRESET_OD	 		BIT6		// 1 = SRESET is open-drain
#define MDCR1_DLL_MAX_DELAY 	BIT2		// 1 = DLL_max_mode, longer DLL delay line length, 0 = normal delay

// PCI8245_MDCR2 - Miscellaneous driver control register 2
#define MDCR2_SDRAM_DSCD_TAP2	0x00 << 4	// SDRAM data in sample clock delay.
#define MDCR2_SDRAM_DSCD_TAP3	0x01 << 4	// See the MPC8245 Hardware Specification
#define MDCR2_SDRAM_DSCD_TAP0	0x02 << 4	// for more information on the
#define MDCR2_SDRAM_DSCD_TAP1	0x03 << 4	// setting of these bits..

// PCI8245_EUMB_BASE - Embedded utilities memory block base address register
#define EUMB_BASE(_x_)			((_x_ & 0xfff) << 20)

// PCI8245_MEM_START1 - Memory starting address register 1
#define MEM_START1_BNK3(_x_)	(((_x_ & 0x0ff00000) >> 20) << 24)
#define MEM_START1_BNK2(_x_)	(((_x_ & 0x0ff00000) >> 20) << 16)
#define MEM_START1_BNK1(_x_)	(((_x_ & 0x0ff00000) >> 20) << 8)
#define MEM_START1_BNK0(_x_)	(((_x_ & 0x0ff00000) >> 20) << 0)

// PCI8245_MEM_START2 - Memory starting address register 1
#define MEM_START2_BNK7(_x_)	(((_x_ & 0x0ff00000) >> 20) << 24)
#define MEM_START2_BNK6(_x_)	(((_x_ & 0x0ff00000) >> 20) << 16)
#define MEM_START2_BNK5(_x_)	(((_x_ & 0x0ff00000) >> 20) << 8)
#define MEM_START2_BNK4(_x_)	(((_x_ & 0x0ff00000) >> 20) << 0)

// PCI8245_XMEM_START1 - Extended memory starting address register 1
#define XMEM_START1_BNK3(_x_)	(((_x_ & 0x70000000) >> 28) << 24)
#define XMEM_START1_BNK2(_x_)	(((_x_ & 0x70000000) >> 28) << 16)
#define XMEM_START1_BNK1(_x_)	(((_x_ & 0x70000000) >> 28) << 8)
#define XMEM_START1_BNK0(_x_)	(((_x_ & 0x70000000) >> 28) << 0)

// PCI8245_XMEM_START2 - Extended memory starting address register 2
#define XMEM_START2_BNK7(_x_)	(((_x_ & 0x70000000) >> 28) << 24)
#define XMEM_START2_BNK6(_x_)	(((_x_ & 0x70000000) >> 28) << 16)
#define XMEM_START2_BNK5(_x_)	(((_x_ & 0x70000000) >> 28) << 8)
#define XMEM_START2_BNK4(_x_)	(((_x_ & 0x70000000) >> 28) << 0)

// PCI8245_MEM_END1 - Memory end address register 1
#define MEM_END1_BNK3(_x_)		(((_x_ & 0x0ff00000) >> 20) << 24)
#define MEM_END1_BNK2(_x_)		(((_x_ & 0x0ff00000) >> 20) << 16)
#define MEM_END1_BNK1(_x_)		(((_x_ & 0x0ff00000) >> 20) << 8)
#define MEM_END1_BNK0(_x_)		(((_x_ & 0x0ff00000) >> 20) << 0)

// PCI8245_MEM_END2 - Memory end address register 1
#define MEM_END2_BNK7(_x_)		(((_x_ & 0x0ff00000) >> 20) << 24)
#define MEM_END2_BNK6(_x_)		(((_x_ & 0x0ff00000) >> 20) << 16)
#define MEM_END2_BNK5(_x_)		(((_x_ & 0x0ff00000) >> 20) << 8)
#define MEM_END2_BNK4(_x_)		(((_x_ & 0x0ff00000) >> 20) << 0)

// PCI8245_XMEM_END1 - Extended memory end address register 1
#define XMEM_END1_BNK3(_x_)		(((_x_ & 0x70000000) >> 28) << 24)
#define XMEM_END1_BNK2(_x_)		(((_x_ & 0x70000000) >> 28) << 16)
#define XMEM_END1_BNK1(_x_)		(((_x_ & 0x70000000) >> 28) << 8)
#define XMEM_END1_BNK0(_x_)		(((_x_ & 0x70000000) >> 28) << 0)

// PCI8245_XMEM_END2 - Extended memory end address register 2
#define XMEM_END2_BNK7(_x_)		(((_x_ & 0x70000000) >> 28) << 24)
#define XMEM_END2_BNK6(_x_)		(((_x_ & 0x70000000) >> 28) << 16)
#define XMEM_END2_BNK5(_x_)		(((_x_ & 0x70000000) >> 28) << 8)
#define XMEM_END2_BNK4(_x_)		(((_x_ & 0x70000000) >> 28) << 0)

// PCI8245_MEM_EN - Memory bank enable register
#define MEM_EN_BNK7				BIT7	// 1 = Enable SDRAM Bank 7
#define MEM_EN_BNK6				BIT6	// 1 = Enable SDRAM Bank 6
#define MEM_EN_BNK5				BIT5	// 1 = Enable SDRAM Bank 5
#define MEM_EN_BNK4				BIT4	// 1 = Enable SDRAM Bank 4
#define MEM_EN_BNK3				BIT3	// 1 = Enable SDRAM Bank 3
#define MEM_EN_BNK2				BIT2	// 1 = Enable SDRAM Bank 2
#define MEM_EN_BNK1				BIT1	// 1 = Enable SDRAM Bank 1
#define MEM_EN_BNK0				BIT0	// 1 = Enable SDRAM Bank 0

// PCI8245_PMOD_CTR	- Page mode counter/timer
#define PGMAX(_x_)		(_x_ & 0xff)	// multiply x 64 to get max page mode open time or Tras.

// PCI8245_PIF_CFG1 - Processor interface configuration 1
#define PIF_CFG1_RCS_PCI		0x00 << 20	// Boot ROM is located on PCI bus
#define PIF_CFG1_RCS_LCL		0x01 << 20	// Boot ROM is located on MPC8245 Local Bus
#define PIF_CFG1_CPU_603		0x02 << 17	// Processor Type - must set this!
#define PIF_CFG1_FLASH_WR_EN	BIT12		// 1 = Flash write is enabled.
#define PIF_CFG1_MCP_EN			BIT11		// 1 = Machine check is enabled.
#define PIF_CFG1_CF_DPARK		BIT9		// 1 = Processor core is parked on data bus. It is recommended to set this bit
#define PIF_CFG1_DEC			BIT8		// 1 = Enable processor core decrementer in extended addressing mode
#define PIF_CFG1_NO_WIDTH_CHECK	BIT7		// 1 = Writes to flash/ROM space are not restricted to bus width writes.
#define PIF_CFG1_ST_GATH_EN 	BIT6		// 1 = Store gathering is enabled.
#define PIF_CFG1_LE_MODE		BIT5		// 1 = Little-endian mode, 0 = Big-endian mode
#define PIF_CFG1_MUST_SET		BIT4		// Reserved and must be set
#define PIF_CFG1_CF_APARK		BIT3		// 1 = Processor core is parked on peripheral address bus, should set
#define PIF_CFG1_SPEC_PCI_RDS	BIT2		// 1 = Speculative reads of Memory by PCI masters are enabled.

// PCI8245_PIF_CFG2 - Processor interface configuration 2
#define PIF_CFG2_SER_ON_CFG 	BIT29		// 1 = Configuration writes to PCI cause MPC8245 to serialize and flush internal buffers.
#define PIF_CFG2_NO_SNOOP_EN	BIT27		// 1 = Snooping is disabled losing cache coherency, but faster performance
#define PIF_CFG2_CF_FF0_LOCAL 	BIT26		// 1 = ROM/Flash remapping enabled. The lower 8 Mbytes of the
											//     ROM/Flash address space are remapped to the processor/memory
											//     bus. ROM/Flash accesses in the range
											//     0xFF00_0000 to 0xFF7F_FFFF are directed to the processor/memory
											//     bus. ROM/Flash accesses in the range
											//     0xFF80_0000 to 0xFFFF_FFFF are directed to the PCI bus.
#define PIF_CFG2_FLASH_WR_LOCK	BIT25		// 1 = Write operations to Flash are disabled.  Cleared only by MPC8245 hard reset.
#define PIF_CFG2_CF_SNOOP_WS_0	0x00 << 18	// Snoop wait states.   These bits control the minimum number of wait
#define PIF_CFG2_CF_SNOOP_WS_1	0x01 << 18  //     states for the address phase in a snoop cycle.
#define PIF_CFG2_CF_SNOOP_WS_2	0x02 << 18  //     1 recommended for 1:1 or 3:2 cpu to mem clock ratios
#define PIF_CFG2_CF_SNOOP_WS_3	0x03 << 18  //     0 recommended for all other ratios
#define PIF_CFG2_CF_APHASE_WS_0 0x00 << 2	// Internal Address phase wait states. These bits control the minimum
#define PIF_CFG2_CF_APHASE_WS_1 0x01 << 2	//   number of address phase wait states (in clock cycles) for
#define PIF_CFG2_CF_APHASE_WS_2 0x02 << 2	//   processor-initiated operations.
#define PIF_CFG2_CF_APHASE_WS_3 0x03 << 2	//   For optimal performance, this parameter should be set to 0

// PCI8245_ERR_EN1 - Error enabling register 1
#define ERR_EN1_PCI_TGT_SERR 	BIT7		// 1 = Report PCI Target SERR Assertions
#define ERR_EN1_PCI_TGT_PERR	BIT6		// 1 = Report PCI MPC8245 as Target can Assert PERR
#define ERR_EN1_MEM_SEL_ERR		BIT5 		// 1 = Report Memory Select Errors
#define ERR_EN1_MEM_REF_ERR		BIT4		// 1 = Report Memory Refresh Overflow Errors
#define ERR_EN1_PCI_MST_PERR	BIT3		// 1 = Report PCI MPC8245 Master PERR Assertions
#define ERR_EN1_MEM_PAR_ERR		BIT2		// 1 = Report Memory Parity (Any) or ECC Errors (above threshold in PCI8245_ECC_ERR_TRG) 
#define ERR_EN1_PCI_MST_ABRT	BIT1		// 1 = Report PCI MPC8245 Master Abort Errors
#define ERR_EN1_CPU_BERR		BIT0		// 1 = Report Processor Bus Errors

// PCI8245_ERR_DET1 - Error detection register 1
#define ERR_DET1_PCI_TGT_SERR 	BIT7		// 1 = PCI Target SERR Asserted
#define ERR_DET1_PCI_TGT_PERR 	BIT6		// 1 = PCI MPC8245 as Target has Asserted PERR
#define ERR_DET1_MEM_SEL_ERR	BIT5		// 1 = Memory Select Error Detected
#define ERR_DET1_MEM_REF_ERR	BIT4		// 1 = Memory Refresh Overflow has Occurred
#define ERR_DET1_CPU_PCI		BIT3		// 0 = Error is on Processor Initiated Cycle, 1 = Error is on PCI Initiated Cycle
#define ERR_DET1_MEM_PAR_ERR	BIT2		// 1 = Memory Read Parity Detected or ECC Single Bit Error Threshhold Exceeded
#define ERR_DET1_CPU_TRANS_ERR	0x01 << 0	// 01 = Unsupported transfer attributes.

// PCI8245_PBUS_ERR Processor internal bus error status
#define PBUS_ERR_TT_MASK		0xF8		// TT0-4 Latched at time of Bus Error
#define PBUS_ERR_TSIZ_MASK		0x07		// TSIZ0-2 Latched at time of Bus Error

// PCI8245_ERR_EN2 - Error enabling register 2
#define ERR_EN2_PCI_ADD_PERR	BIT7		// 1 = PCI Address Parity Errors Generate *MCP
#define ERR_EN2_PCI_SERR		BIT6		// 1 = SERR (MPC8245 or PCI Initiated Cycles) Detection Enabled
#define ERR_EN2_MEM_ECC_ERR		BIT3		// 1 = ECC Multibit Error Detection Enabled
#define ERR_EN2_MEM_WR_ERR		BIT2		// 1 = CPU Memory Write Parity Error Detection Enabled
#define ERR_EN2_PCI_TGT_ABRT	BIT1		// 1 = Target Abort Error Detection Enabled
#define ERR_EN2_FLASH_WR_ERR	BIT0		// 1 = Flash Write Error Enabled when PIF_CFG1_FLASH_WR_EN = 0 or PIF_CFG2_FLASH_WR_LOCK = 1

// PCI8245_ERR_DET2 - Error detection register 2
#define ERR_DET2_BAD_ADD_ERR	BIT7		// 1 = The address in ERR_ADD is not valid.
#define ERR_DET2_PCI_SERR		BIT6		// 1 = SERR detected
#define ERR_DET2_MEM_ECC_ERR	BIT3		// 1 = ECC Multibit Error Detected
#define ERR_DET2_MEM_WR_ERR		BIT1		// 1 = Processor Memory Write Parity Error Detected
#define ERR_DET2_FLASH_WR_ERR	BIT0		// 1 = Flash Write Attempt when PIF_CFG1_FLASH_WR_EN = 0 or PIF_CFG2_FLASH_WR_LOCK = 1

// PCI8245_PCI_ERR - PCI bus error status register
#define PCI_ERR_CPU_STAT		BIT4		// 0 = MPC8245 was the PCI master, 1 = MPC8245 was the PCI target.
#define PCI_ERR_CBE_MASK		0x0f		// *C/BE Signals latched at the time of the error

// PCI8245_XROM_CFG1 and CFG2 - Extended ROM configuration registers 1 (RCS2) and 2 (RCS3)
#define XROM_CFG12_EN			BIT31		// 1 = RCS2 enabled
#define XROM_CFG12_BURST		BIT30		// 1 = RCS uses burst-mode timing. Use RCS2_ROMNAL for burst beats.
#define XROM_CFG12_DBW_8		0x00 << 28	// RCS2 Bus Width = 8-bit data path with gathering
#define XROM_CFG12_DBW_16		0x01 << 28	// RCS2 Bus Width = 16-bit data path with gathering
#define XROM_CFG12_DBW_32		0x02 << 28	// RCS2 Bus Width = 32-bit data path with gathering
#define XROM_CFG12_DBW_64		0x03 << 28	// RCS2 Bus Width = 64-bit if DBUS0 = 1, 32-bit if DBUS0 = 0
#define XROM_CFG12_CTL_IND		0x00 << 26	// Independent Timing Mode, uses timing from this register.
#define XROM_CFG12_CTL_BASE 	0x01 << 26  // ROM/FLASH Timing Mode, uses timing from MCCRs.
#define XROM_CFG12_CTL_XSTB		0x02 << 26	// PortX strobe mode, Note that flash recovery time is disabled in this mode
#define XROM_CFG12_CTL_XHSK		0x03 << 26	// PortX handshake mode, Note that flash recovery time is disabled in this mode
#define XROM_CFG12_ROMFAL(_x_)	((_x_ & 0x1f) << 20)	// x + 3 = Nonburst Total Access Time, or Burst First Access Time
														// for reads, for Writes x = Write Pulse Low Time
#define XROM_CFG12_ROMNAL(_x_)	((_x_ & 0x0f) << 15)	// x + 3 = Next Access latency for Reads, for Writes x + 4
#define XROM_CFG12_ASFAL(_x_)	((_x_ & 0x1f) << 10)	// x = falling edge AS relative to falling edge of RCS2 (Port X only)
#define XROM_CFG12_ASRISE(_x_)	((_x_ & 0x1f) << 5)		// x = rising edge AS relative to falling edge of RCS2 (Port X only)
#define XROM_CFG12_TSWAIT(_x_)	((_x_ & 0x1f) << 0)		// x = wait states before subsequent transaction starts

// PCI8245_XROM_CFG3 and 4 - Extended ROM configuration registers 3 (RCS2) and 4 (RCS3)
#define XROM_CFG34_SADDR(_x_)	(_x_ & 0x0ffff000)		// Starting Address Mask - Upper Nibble is ignored and is set to 0x7 by CPU
#define XROM_CFG34_SIZE_4K		0x00 << 0		// Size of RCS Bank
#define XROM_CFG34_SIZE_8K		0x01 << 0
#define XROM_CFG34_SIZE_16K		0x02 << 0
#define XROM_CFG34_SIZE_32K		0x03 << 0
#define XROM_CFG34_SIZE_64K		0x04 << 0
#define XROM_CFG34_SIZE_128K	0x05 << 0
#define XROM_CFG34_SIZE_256K	0x06 << 0
#define XROM_CFG34_SIZE_512K	0x07 << 0
#define XROM_CFG34_SIZE_1M		0x08 << 0
#define XROM_CFG34_SIZE_2M		0x09 << 0
#define XROM_CFG34_SIZE_4M		0x0a << 0
#define XROM_CFG34_SIZE_8M		0x0b << 0
#define XROM_CFG34_SIZE_16M		0x0c << 0
#define XROM_CFG34_SIZE_32M		0x0d << 0
#define XROM_CFG34_SIZE_64M		0x0e << 0
#define XROM_CFG34_SIZE_128M	0x0f << 0

// PCI8245_MAPB_OPT -  Address map B options register
#define MAPB_OPT_CPU_FD_ALIAS 	BIT7	// 0 = Access are routed normally, 1 = CPU accesses at 0xFDxx_xxxx are forwarded PCI MEM 0x00xx_xxxx
#define MAPB_OPT_PCI_FD_ALIAS	BIT6	// 0 = No response, 1 = MPC8245, as PCI target, forwards PCI 0xFDxx_xxxx to Memory at 0x00xx_xxxx
#define MAPB_OPT_DLL_RESET		BIT5	// Write 1, then 0 to reset the DLL tap point.  Must be done during init.
#define MAPB_OPT_PCI_HOLE		BIT3	// 0 = MPC8245, forwards PCI 0x000A_0000 - 0x000F_FFFF to memory, 1 = no response
#define MAPB_OPT_CPU_HOLE		BIT2	// 0 = MPC8245, forwards CPU 0x000A_0000 - 0x000B_FFFF to memory, 1 = forwards to PCI

// PCI8245_MCCR1 - Memory Control/Configuration register 1
#define MCCR1_ROMNAL(_x_)		((_x_ & 0x0f) << 28)	// For reads, x + 3 = Next Access Time.  For writes x + 4 = Write Recovery Time
#define MCCR1_ROMFAL(_x_)		((_x_ & 0x1f) << 23)	// For reads, x + 3 = Access Time.  For writes x + 2 = Write Pulse Low Time
#define MCCR1_BURST 			BIT20	// 1 = Burst-mode ROM access timing
#define MCCR1_MEMGO				BIT19	// 1 = RAM Enable. Note that this bit must be set after all other configuration parameters
#define MCCR1_SREN				BIT18	// 1 = Enables the SDRAM self refresh during sleep mode
#define MCCR1_SDRAM_EN			BIT17	// 0 = SDRAM Enable, must be cleared to 0 before setting MCCR1_MEMGO
#define MCCR1_PCKEN				BIT16	// 1 = Enables parity for registered mode memory only (must be clear for inline mode)
#define MCCR1_BNK_R12B4(_b_)	0x00 << ((_b_ & 0x07) << 1)	// Bank 0-7 row size = 12 or 11 bits, 4 banks
#define MCCR1_BNK_R13B4(_b_)	0x10 << ((_b_ & 0x07) << 1)	// Bank 0-7 row size = 13 bits, 4 banks
#define MCCR1_BNK_R11B2(_b_)	0x11 << ((_b_ & 0x07) << 1)	// Bank 0-7 row size = 12 bits, 2 banks

// PCI8245_MCCR2 - Memory Control/Configuration register 2
#define MCCR2_TS_WAIT(_x_)		((_x_ & 0x07) << 29)	// Transaction start wait states timer
#define MCCR2_ASRISE(_x_)		((_x_ & 0x0f) << 25)	// x = rising edge AS relative to falling edge (Port X mode only)
#define MCCR2_ASFAL(_x_)		((_x_ & 0x0f) << 21)	// x = falling edge AS relative to falling edge (Port X mode only)
#define MCCR2_INLINE_PAR		BIT20	// 0 = Use ECC, 1 = Use parity (if MCCR2_INLINE_WR_PAR or MCCR2_INLINE_RD_PAR set)
#define MCCR2_INLINE_WR_PAR		BIT19	// 1 = Enable Parity Checking/Reporting on Memory Writes in inline mode only
#define MCCR2_INLINE_RD_PAR		BIT18	// 1 = Enable Parity Checking/Reporting on Memory Reads in inline mode only
#define MCCR2_REFINT(_x_)		((_x_ & 0x3fff) << 2)	// x = number of clock cycles between CBR refresh cycles
#define MCCR2_RSV_PG			BIT1	// 0 = Four open page mode (default), 1 = Reserve one of four page registers
#define MCCR2_RMW_PAR 			BIT0 	// 1 = RMW parity enabled. Must be set when MCCR4_BUF_TYPE = 0x01 and MCCR2_INLINE_PAR = 0)

// PCI8245_MCCR3 - Memory Control/Configuration register 3
#define MCCR3_BSTOPRE_2_5(_x_)	(((_x_ >> 2) & 0x0f) << 28)	// get bits 2-5 of the Burst to Precharge value.  BSTOPRE = open page time
#define MCCR3_REFREC(_x_)		((_x_ & 0x0f) << 24)		// refresh to activate time, (note: 0 = 16 clocks)

// PCI8245_MCCR4 - Memory Control/Configuration register 4
#define MCCR4_PRETOACT(_x_)		((_x_ & 0x0f) << 28)		// Precharge to activate time, (note: 0 = 16 clocks)
#define MCCR4_ACTOPRE(_x_)		((_x_ & 0x0f) << 24)		// Activate to precharge time, (note: 0 = 16 clocks)
#define MCCR4_WMODE				BIT23						// Length of burst for 32-bit data. 0 4 beats per burst, 1 = 8 beats
#define MCCR4_BUF_TYPE_INLINE	BIT22						// MSB of Memory buffer type field
#define MCCR4_BUF_TYPE_REG		BIT20 						// LSB of memory buffer type field
#define MCCR4_EXTROM			BIT21						// 1 = Extended 256 Mbytes ROM space enabled
#define MCCR4_BSTOPRE_0_1(_x_)	(((_x_ >> 0) & 0x03) << 18)	// Bits 0-1 of the Burst to Precharge value.  BSTOPRE = open page time
#define MCCR4_REGDIMM			BIT15						// 1 = Registered DIMMs selected
#define MCCR4_SMODE_CAS_L1		0x01 << 12					// CAS Latency to be written to SDRAM Mode Register
#define MCCR4_SMODE_CAS_L2		0x02 << 12
#define MCCR4_SMODE_CAS_L3		0x03 << 12
#define MCCR4_SMODE_SEQ			0x00 << 11					// Burst Wrap Type Sequential - Recommended
#define MCCR4_SMODE_INTER		0x01 << 11					// Burst Wrap Type Interleaved
#define MCCR4_SMODE_BRST_4		0x02 << 8					// Burst Length = 4
#define MCCR4_SMODE_BRST_8		0x03 << 8					// Burst Length = 8
#define MCCR4_ACTORW(_x_)		((_x_ & 0x0f) << 4)  		// Activate to read/write time, (note: 0 = 16 clocks)
#define MCCR4_BSTOPRE_6_9(_x_)	(((_x_ >> 6) & 0x0f) << 0)	// Bits 6-9 of the Burst to Precharge value.  BSTOPRE = open page time


// Define some default values for the memory controller
// that need to be set in the reset init code.
// note: all wait states and timing are defined in units of
// MEMCLK, which is 83Mhz (12ns) for the CSB281

// BSTOPRE defines how long the 8245 will leave a page open before
// issuing a precharge command.  Too low and we close pages earlier
// than needed, too high and we always wait for the page to close 
// before we can begin another cycle.  300 clocks equals about 100
// instructions on average and seems sensible.  But hey, who the
// heck really knows? 
#define BSTOPRE		300	// 300 * 12ns = 3.6us

// PGMAX is the maximum time the 8245 will wait before closing
// a page even if we've been hitting the page all along (unlike
// BSTOPRE which will be reset every time we get a page hit). 
// PGMAX must take into account the fact that we may get stalled
// by an S1D13506 variable latency access.  To be safe we assume
// that could be up to 50usec, so subtract that from 100usec (which
// is the maximmum defined by the SDRAM devices) to get 50usec. 
// Divide that by 12ns to get 4166 clocks.  Divide by 64 to get the
// value we put into PCI8245_PGMAX or 65.
#define	PGMAX_DEF 	65

// remember to set MCCR1_DEF0 first, the MCCR2-4 and finally MCCR1_DEF1 last
// MCCR1_ROMFAL -    (3 + 10) * 12ns = 156ns
// MCCR1_ROMNAL -    (3 + 0) * 12ns = 36ns
// MCCR1_MEMGO RAM - Enable must be set after all other configuration parameters
// MCCR1_R12B4_BNK - bank 0 is 12 row, 4 bank SDRAM
#define MCCR1_DEF0	(MCCR1_ROMFAL(16) | MCCR1_ROMNAL(8) | MCCR1_BNK_R12B4(0))			
#define MCCR1_DEF1	(MCCR1_ROMFAL(16) | MCCR1_ROMNAL(8) | MCCR1_MEMGO | MCCR1_BNK_R12B4(0))			

// MCCR2_TS_WAIT - (2 + 0) * 12ns = 24ns
// MCCR2_ASFAL -   0
// MCCR2_ASRISE -  MCCR1_ROMFAL - 1, we use *AS to do writes to the BCSR
// MCCR2_REFINT -  1200 * 12ns = 14usec refresh cycle
#define MCCR2_DEF	(MCCR2_TS_WAIT(0) | MCCR2_ASFAL(0) | MCCR2_ASRISE(15) | MCCR2_REFINT(143))  		

// MCCR3_REFREC - 6 * 12 = 72ns Trfc
#define MCCR3_DEF	(MCCR3_BSTOPRE_2_5(BSTOPRE) | MCCR3_REFREC(6))

// MCCR4_PRETOACT(3) -  36ns
// MCCR4_ACTOPRE(6) -   2 clocks plus # of beats per burst = 6
// MCCR4_BUF_TYPE_REG - Registered Mode
// MCCR4_EXTROM -       Enable Extended ROM Space
// MCCR4_BSTOPRE_0_1 -  Bits 0-1 of the Burst to Precharge value.  BSTOPRE = open page time
// MCCR4_SMODE_CAS_L3 - CAS Latency 3
// MCCR4_SMODE_SEQ -    Burst Wrap Type Sequential - Recommended
// MCCR4_SMODE_BRST_4 - Burst Length = 4
// MCCR4_ACTORW(3) -    Tell CPU we are using 3 Clock CAS Latency
// MCCR4_BSTOPRE_6_9 -  Bits 6-9 of the Burst to Precharge value.  BSTOPRE = open page time
#define MCCR4_DEF	(MCCR4_PRETOACT(3) | MCCR4_ACTOPRE(5) | MCCR4_BUF_TYPE_REG | MCCR4_EXTROM \
				  |	MCCR4_BSTOPRE_0_1(BSTOPRE) | MCCR4_SMODE_CAS_L3 | MCCR4_SMODE_SEQ \
				  |	MCCR4_SMODE_BRST_4 | MCCR4_ACTORW(3) | MCCR4_BSTOPRE_6_9(BSTOPRE))


#define MEM_START1_DEF	MEM_START1_BNK0(0x00000000)	// SDRAM is 0x000_0000 to 0x03FF_FFFF	- 64Mbytes
#define MEM_END1_DEF	MEM_END1_BNK0(0x03fffffff)
#define MEM_EN_DEF		MEM_EN_BNK0					// Enable SDRAM Bank only



