//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         05-26-2002
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to each CPU/Board combination
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "lh79520_uart.h"
#include "lh79520.h"
#include "cpuio.h"
#include "warmstart.h"

extern ulong i2c_init();

uchar	bcsr_led_copy;		// BCSR LED register is write only
							// (reading gets the switches)

/* InitUART():
 * This is the top-level UART initialization for the monitor. 
 */
void
InitUART(int baud)
{
	// Disable first
	UART0_REG(UART_LCR_L) = 0x00000000;

	// clear any pending interrupts and mask all further interrupts
	UART0_REG(UART_ICR) = UART_INT_ALL;
	UART0_REG(UART_IMSC) = 0x00000000;

	// The LH79520 derives it's baud clock from the 14.7456Mhz input clock
	UART0_REG(UART_IBRD) = ((14745600/16)/baud);	// integer divisor
	UART0_REG(UART_FBRD) = 0x00000000;				// fractional divisor

	// set the port to 8-bits, 1 stop - This must be done after the baud rate!
	UART0_REG(UART_LCR_H) = UART_LCR_H_LEN8 | UART_LCR_H_FEN;

	// Enable
	UART0_REG(UART_LCR_L) = (UART_LCR_L_EN | UART_LCR_L_TXE | UART_LCR_L_RXE);
}

int
ConsoleBaudSet(int baud)
{
	// The LH79520 derives it's baud clock from the 14.7456Mhz input clock
	UART0_REG(UART_IBRD) = ((14745600/16)/baud);	// integer divisor
	UART0_REG(UART_FBRD) = 0x00000000;				// fractional divisor
	return(0);
}

int
devInit(int baud)
{
	extern int fs6377_init(int);

	// Turn on User LED 0
	bcsr_led_copy = BCSR_LED0;
	BCSR_REG(BCSR_LED) = bcsr_led_copy;

	// Initialize the UART
    InitUART(baud);

	// Initialize the I2C Port
	i2c_init();

	// initialize the I2C PLL
	fs6377_init(0);

	return 0;
}

/* target_console_empty():
 * Target-specific portion of flush_console() in chario.c.
 * This function returns 1 if there are no characters waiting to
 * be put out on the UART; else return 0 indicating that the UART
 * is still busy outputting characters from its FIFO.
 */
int
target_console_empty(void)
{
	if (!(UART0_REG(UART_FR) & UART_FR_TXFF))
		return(1);
	return(0);
}
	
/* target_putchar():
 */
int
target_putchar(char c)
{
    /* Wait for transmit ready bit */
    while(1) {
        if (UART0_REG(UART_FR) & UART_FR_TXFE)
            break;
    }

	UART0_REG(UART_DR) = c;
    return((int)c);
}

/* gotachar():
 * Return 0 if no char is avaialable at UART rcv fifo; else 1.
 * Do NOT pull character out of fifo, just return status. 
 */
int
target_gotachar()
{
    if (UART0_REG(UART_FR) & UART_FR_RXFE)
        return(0);
    return(1);
}

/* target_getchar():
 */
int 
target_getchar()
{
    char    c;

    /* Retrieve character */
    c = UART0_REG(UART_DR);

    return((int)c);
}

void
initCPUio()
{
}

void
target_reset(void)
{
	/* Should be replaced with something that touches hardware to
	 * cause a more robust reset.
	 */
	warmstart(INITIALIZE);
}

extern void putpsr(ulong);
extern ulong getpsr(void);

void
intsrestore(ulong psr)
{
    putpsr(psr);
}

/*
 * Read the program status register (CPSR)
 * and set the FIQ and IRQ bits.
 */
ulong
intsoff(void)
{
    ulong  psr;

    psr = getpsr();

    /*
     * Set bit 6, bit 7 to disable interrupts.
     */
    putpsr(psr | 0x000000c0);
    return(psr);
}

/* cacheInitForTarget():
    Enable instruction cache only...
*/
void
cacheInitForTarget()
{
    asm("   MRC p15, 0, r0, c1, c0, 0");
    asm("   ORR r0, r0, #0x1000");  /* bit 12 is ICACHE enable*/
    asm("   MCR p15, 0, r0, c1, c0, 0");

    /* Flush instruction cache */
    asm("   MCR p15, 0, r0, c7, c5, 0");
}
