/*********************************************************************
 *
 * Copyright:
 *  MOTOROLA, INC. All Rights Reserved.  
 *  You are hereby granted a copyright license to use, modify, and
 *  distribute the SOFTWARE so long as this entire notice is
 *  retained without alteration in any modified and/or redistributed
 *  versions, and that such modified versions are clearly identified
 *  as such. No licenses are granted by implication, estoppel or
 *  otherwise under any patents or trademarks of Motorola, Inc. This 
 *  software is provided on an "AS IS" basis and without warranty.
 *
 *  To the maximum extent permitted by applicable law, MOTOROLA 
 *  DISCLAIMS ALL WARRANTIES WHETHER EXPRESS OR IMPLIED, INCLUDING 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR
 *  PURPOSE AND ANY WARRANTY AGAINST INFRINGEMENT WITH REGARD TO THE 
 *  SOFTWARE (INCLUDING ANY MODIFIED VERSIONS THEREOF) AND ANY 
 *  ACCOMPANYING WRITTEN MATERIALS.
 * 
 *  To the maximum extent permitted by applicable law, IN NO EVENT
 *  SHALL MOTOROLA BE LIABLE FOR ANY DAMAGES WHATSOEVER (INCLUDING 
 *  WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS 
 *  INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR OTHER PECUNIARY
 *  LOSS) ARISING OF THE USE OR INABILITY TO USE THE SOFTWARE.   
 * 
 *  Motorola assumes no responsibility for the maintenance and support
 *  of this software
 ********************************************************************/

#include "config.h"
#include "cpuio.h"
#include "ether.h"
#include "genlib.h"
#include "stddefs.h"
#include "warmstart.h"
#include "csb360_gpio.h"
#include "mcf5272.h"

void    mcf5272_sim_init (MCF5272_IMM *);
void    mcf5272_gpio_init (MCF5272_IMM *);
void    mcf5272_cs_init (MCF5272_IMM *);
void    mcf5272_uart_init(MCF5272_IMM *);

/* Delay for some usecs using Timer 3.
 * resolution is 10 usec using a prescale
 * value of 42 and the CLK/16 option.  
 * This limits the maximum time to 650 msecs
 */
void
udelay(int delay)
{
    MCF5272_IMM *imm;

    imm = mcf5272_get_immp();

	/* minimum delay is 10 usec
	 */
	if (delay < 10)
		delay = 10;
	else if (delay > 650000)
		delay = 650000;

	delay = delay/10;

	/* set the desired usecs into the timer ref register
	 */
	MCF5272_WR_TIMER_TRR(imm, 3, delay);

	/* Reset, then enable Timer 3
	 */
	MCF5272_WR_TIMER3_TMR(imm, 0x0000);	
	MCF5272_WR_TIMER3_TMR(imm, MCF5272_TIMER_TMR_PS(42)
						     | MCF5272_TIMER_TMR_FRR
							 | MCF5272_TIMER_TMR_CLK_DIV16
							 | MCF5272_TIMER_TMR_RST);

	/* wait for the reference to be reached
	 */
	while(1) {
		if (MCF5272_RD_TIMER_TER(imm,3) & MCF5272_TIMER_TER_REF);
			break;
	}
}

/********************************************************************/

/********************************************************************/

/*
 * Out of reset, the low-level assembly code calls this routine to
 * obtain the address at which to map the MBAR.
 */
void *
mcf5272_mbar (void)
{
    return (void *)IMM_ADDRESS;
}

/********************************************************************/

/*
 * Out of reset, the low-level assembly code calls this routine to
 * obtain the address at which to map the RAMBAR.
 */
void *
mcf5272_rambar (void)
{
    return (void *)INT_SRAM_ADDRESS;
}

/********************************************************************/

/*
 * Out of reset, the low-level assembly code calls this routine to
 * initialize the MCF5272 for this board. A temporary stack has been
 * setup in the internal SRAM, and the stack pointer will be changed
 * to point to SDRAM once this routine returns.
 */
void
mcf5272_init (MCF5272_IMM *imm)
{
	volatile long *lp, l;

    mcf5272_uart_init(imm);
	mcf5272_sim_init(imm);
	mcf5272_gpio_init(imm);
    mcf5272_cs_init(imm);

    /* Turn Instruction Cache ON
    mcf5272_wr_cacr(0x81000500);
     */

	/* Wake up the DRAM
	 */
	for(lp=SDRAM_ADDRESS;lp<(long *)(SDRAM_ADDRESS+0x10000);lp++)
		l = *lp;
}

/****************************************************************/
void
mcf5272_sim_init (MCF5272_IMM *imm)
{
    /*******************************************************
    *
    * This routine executed upon reset to initialize
    * MCF5272 SIM registers.  An initial stack must have been setup
    * in the SRAM in order for this routine to work.
    *
    *******************************************************/

    /*
     * Initialize System Config Register
     * Setup Watch Dog Timeout
     */
    MCF5272_WR_SIM_SCR(imm,MCF5272_SIM_SCR_HWWD_1024);

    /*
     * Initialize System Protection Register
     * Enable all bus error exceptions
     */
#if 0
    MCF5272_WR_SIM_SPR(imm,0
        | MCF5272_SIM_SPR_ADC | MCF5272_SIM_SPR_ADCEN
        | MCF5272_SIM_SPR_WPV | MCF5272_SIM_SPR_WPVEN
        | MCF5272_SIM_SPR_SMV | MCF5272_SIM_SPR_SMVEN
        | MCF5272_SIM_SPR_SBE | MCF5272_SIM_SPR_SBEEN
        | MCF5272_SIM_SPR_HWT | MCF5272_SIM_SPR_HWTEN
        | MCF5272_SIM_SPR_RPV | MCF5272_SIM_SPR_RPVEN
        | MCF5272_SIM_SPR_EXT | MCF5272_SIM_SPR_EXTEN
        | MCF5272_SIM_SPR_SUV | MCF5272_SIM_SPR_SUVEN
        ) ;
#endif
    /* 
     * Initilize Interrupt Control Registers
     * Mask all interrupt sources
     */
    MCF5272_WR_SIM_ICR1(imm,0x88888888);
    MCF5272_WR_SIM_ICR2(imm,0x88888888);
    MCF5272_WR_SIM_ICR3(imm,0x88888888);
    MCF5272_WR_SIM_ICR4(imm,0x88880000);

    /* 
     * Initialize Programmable Interrupt Transition Register
     * Make all external interrupts trigger on falling edge
     */ 
    MCF5272_WR_SIM_PITR(imm,MCF5272_SIM_PITR_NEG_EDGE);

    /*
     * Initialize Programmable Interrupt Vector Register
     * Place peripheral interrupt vectors beginning at vector 64
     */
    MCF5272_WR_SIM_PIVR(imm,MCF5272_SIM_PIVR_NORMAL);

}

/********************************************************************/
void 
mcf5272_gpio_init (MCF5272_IMM *imm)
{
    /*
     * Inititialize Port A
     */

    /* Enable PA11 as QSPI_CS1 signals */
    MCF5272_WR_GPIO_PACNT(imm,0x00800000);

    /* Configure all pins as inputs */
    MCF5272_WR_GPIO_PADDR(imm,0);

    /*
     * Inititialize Port B
     */

    /* Enable UART0 and FEC signals 
       PB6 and PB7 are SCL and SDA
    */
    MCF5272_WR_GPIO_PBCNT(imm,0x55550555);

    /* Configure all GPIO pins as inputs */
    MCF5272_WR_GPIO_PBDDR(imm,0);
    
    /*
     * Inititialize Port C
     * Since we are in 32-bit mode, Port C is not available!
     */

    /*
     * Inititialize Port D
     * Enable UART1 and *INT4.  Leave PWM0-2 in high impedance
     */
    MCF5272_WR_GPIO_PDCNT(imm,0x00000eaa);
}

/********************************************************************/
void
mcf5272_uart_init(MCF5272_IMM *imm)
{
    uint16 ubgs;

    /* Initialize UART0 */
    
    /* Reset Transmitter */
    MCF5272_WR_UART0_UCR(imm,MCF5272_UART_UCR_RESET_TX);

    /* Reset Receiver */
    MCF5272_WR_UART0_UCR(imm,MCF5272_UART_UCR_RESET_RX);

    /* Reset Mode Register */
    MCF5272_WR_UART0_UCR(imm,MCF5272_UART_UCR_RESET_MR);

    /* No parity, 8-bits per character */
    MCF5272_WR_UART0_UMR(imm, 0
        | MCF5272_UART_UMR1_PM_NONE
        | MCF5272_UART_UMR1_BC_8 );

    /* No echo or loopback, 1 stop bit */
    MCF5272_WR_UART0_UMR(imm, 0
        | MCF5272_UART_UMR2_CM_NORMAL
        | MCF5272_UART_UMR2_STOP_BITS_1);

    /* Set Rx and Tx baud by timer */
    MCF5272_WR_UART0_UCSR(imm, 0
        | MCF5272_UART_UCSR_RCS(0xD)
        | MCF5272_UART_UCSR_TCS(0xD));

    /* Mask all USART interrupts */
    MCF5272_WR_UART0_UIMR(imm, 0);
                 
    /* Calculate baud settings */
    ubgs = (uint16)((SYSTEM_CLOCK*1000000)/(UART_BAUD * 32));

    MCF5272_WR_UART0_UBG1(imm, (uint8)((ubgs & 0xFF00) >> 8));
    MCF5272_WR_UART0_UBG2(imm, (uint8)(ubgs & 0x00FF));

    /* Enable receiver and transmitter */
    MCF5272_WR_UART0_UCR(imm, 0
        | MCF5272_UART_UCR_TX_ENABLED
        | MCF5272_UART_UCR_RX_ENABLED);

    /* Disable UART1 */
    MCF5272_WR_UART1_UCR(imm, 0
        | MCF5272_UART_UCR_TX_DISABLED
        | MCF5272_UART_UCR_RX_DISABLED);
}

/********************************************************************/
void
mcf5272_cs_init (MCF5272_IMM *imm)
{
    /* This routine initializes ChipSelects to setup memory devices */
    /* ChipSelect 0 - 8MB or 16MB StrataFLASH, we configure for a
	 * full 16Mg here...
	 */
    MCF5272_WR_CS_CSBR0(imm, MCF5272_CS_BR_BASE(FLASH_BANK0_BASE_ADDR)
        | MCF5272_CS_BR_SRAM
        | MCF5272_CS_BR_PS_16
        | MCF5272_CS_BR_EN);    

    MCF5272_WR_CS_CSOR0(imm, MCF5272_CS_OR_MASK_16M | MCF5272_CS_OR_WS(10));

    /* Chip Select 1 - S1D13706 LCD Controller */
    MCF5272_WR_CS_CSBR1(imm, MCF5272_CS_BR_BASE(S1D13706_ADDRESS)
        | MCF5272_CS_BR_SRAM
        | MCF5272_CS_BR_PS_32
		| MCF5272_CS_OR_ASET	// delay CS, RD and WR one clk after addr
		| MCF5272_CS_OR_RDAH	// Hold addr one clk after CS and RD
		| MCF5272_CS_OR_WRAH	// Hold addr and data one clk after CS and WR
        | MCF5272_CS_BR_EN);

    MCF5272_WR_CS_CSOR1(imm, MCF5272_CS_OR_MASK_512K
        | MCF5272_CS_OR_WS(0x1F));	// external *TA

    /* ChipSelect 7 - 32MB SDRAM */
    MCF5272_WR_CS_CSBR7(imm, MCF5272_CS_BR_BASE(SDRAM_ADDRESS)
        | MCF5272_CS_BR_SDRAM 
        | MCF5272_CS_BR_PS_LINE 
        | MCF5272_CS_BR_EN);

    MCF5272_WR_CS_CSOR7(imm, MCF5272_CS_OR_MASK_32M
        | MCF5272_CS_OR_WS(0x1F));

    /* Do not initialize SDRAM if already running in SDRAM */
    if (!(MCF5272_RD_SDRAMC_SDCCR(imm) & MCF5272_SDRAMC_SDCCR_ACT))
    {
		MCF5272_WR_SDRAMC_SDCTR(imm, MCF5272_SDRAMC_SDCTR_RTP_66MHz
			| MCF5272_SDRAMC_SDCTR_RC(1) | MCF5272_SDRAMC_SDCTR_RP(3)
			| MCF5272_SDRAMC_SDCTR_RCD(2) | MCF5272_SDRAMC_SDCTR_CLT_2);

        MCF5272_WR_SDRAMC_SDCCR(imm, MCF5272_SDRAMC_SDCCR_MCAS_A10
            | MCF5272_SDRAMC_SDCCR_BALOC_A23
            | MCF5272_SDRAMC_SDCCR_REG | MCF5272_SDRAMC_SDCCR_INIT);
    }

    /* ChipSelect 2 to 6 - Empty */
    MCF5272_WR_CS_CSBR2(imm, 0);
    MCF5272_WR_CS_CSOR2(imm, 0);
    MCF5272_WR_CS_CSBR3(imm, 0);
    MCF5272_WR_CS_CSOR3(imm, 0);
    MCF5272_WR_CS_CSBR4(imm, 0);
    MCF5272_WR_CS_CSOR4(imm, 0);
    MCF5272_WR_CS_CSBR5(imm, 0);
    MCF5272_WR_CS_CSOR5(imm, 0);
    MCF5272_WR_CS_CSBR6(imm, 0);
    MCF5272_WR_CS_CSOR6(imm, 0);
}

void
InitUART(int baud)
{
}

void
initCPUio()
{
	mcf5272_init((MCF5272_IMM *)IMM_ADDRESS);
}

void
ledloop1(void)
{
	extern	void GPIO_out(), GPIO_clr(), GPIO_set();
	volatile int delay, count;

	GPIO_out(GPIO_LED0);

	for(count=0;count<3;count++) {
		GPIO_clr(GPIO_LED0);
		for(delay=0;delay<100000;delay++);
		GPIO_set(GPIO_LED0);
		for(delay=0;delay<100000;delay++);
	}
}

void
ledloop2()
{
	extern	void GPIO_out(), GPIO_clr(), GPIO_set();
	volatile int delay, count;

	GPIO_out(GPIO_LED1);

	for(count=0;count<3;count++) {
		GPIO_clr(GPIO_LED1);
		for(delay=0;delay<100000;delay++);
		GPIO_set(GPIO_LED1);
		for(delay=0;delay<100000;delay++);
	}
}

void
ledloop3()
{
	extern	void GPIO_out(), GPIO_clr(), GPIO_set();
	volatile int delay, count;

	GPIO_out(GPIO_LED0);
	GPIO_out(GPIO_LED1);

	for(count=0;count<3;count++) {
		GPIO_clr(GPIO_LED0);
		GPIO_clr(GPIO_LED1);
		for(delay=0;delay<100000;delay++);
		GPIO_set(GPIO_LED0);
		GPIO_set(GPIO_LED1);
		for(delay=0;delay<100000;delay++);
	}
}


int
devInit(int baud)
{
	extern	void GPIO_out(), GPIO_clr(), i2c_init();
	extern	int fs6377_init();

	// turn on debug LED0
	GPIO_out(GPIO_LED0);
	GPIO_clr(GPIO_LED0);

	// initialize the I2C interface
	i2c_init();

	// initialize the Programmable PLL (via I2C)
	if (fs6377_init(0) == -1)
	{	// turn on debug LED1
	 	GPIO_out(GPIO_LED1);
		GPIO_clr(GPIO_LED1);
	}

    return(0);
}


int
target_getchar (void)
{
    MCF5272_IMM *imm;

    imm = mcf5272_get_immp();
    return MCF5272_RD_UART0_URB(imm);
}

int
target_putchar(char ch)
{
    MCF5272_IMM *imm;

    imm = mcf5272_get_immp();

    /* Wait until space is available in the FIFO */
    while (!(MCF5272_RD_UART0_USR(imm) & MCF5272_UART_USR_TXRDY));

    /* Send the character */
    MCF5272_WR_UART0_UTB(imm, (uint8)ch);

    return(ch);
}

int
target_gotachar (void)
{
    MCF5272_IMM *imm;

    imm = mcf5272_get_immp();
    return (MCF5272_RD_UART0_USR(imm) & MCF5272_UART_USR_RXRDY);
}

int
target_console_empty(void)
{
   	if (MCF5272_RD_UART0_USR(IMM_ADDRESS) & MCF5272_UART_USR_TXEMP)
		return(1);
	return(0);
}

int
ConsoleBaudSet(int baud)
{
    uint16 ubgs;
    MCF5272_IMM *imm;

    imm = mcf5272_get_immp();

    /* Calculate baud settings */
    ubgs = (uint16)((SYSTEM_CLOCK*1000000)/(baud * 32));

    MCF5272_WR_UART0_UBG1(imm, (uint8)((ubgs & 0xFF00) >> 8));
    MCF5272_WR_UART0_UBG2(imm, (uint8)(ubgs & 0x00FF));
	return(0);
}

void
target_reset(void)
{
	warmstart(INITIALIZE);
}

void
intsrestore(ulong sr)
{
	extern void putsr(ulong);
	putsr(sr);
}

ulong
intsoff()
{
	extern void putsr(ulong);
	extern ulong getsr(void);
	ushort  sr;

	sr = getsr();
	putsr(sr|0x0700);
	return((int)sr);
}

void
cacheInitForTarget()
{
	extern void mcf5272_wr_cacr();

	/* Enable instruction cache. */
	mcf5272_wr_cacr(0x81000500);
}

