//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO for CSB437TL
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         07/31/2004
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to each CPU/Board combination
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "at91rm9200.h"
#include "at91rm9200_dbgu.h"
#include "at91rm9200_gpio.h"
#include "at91rm9200_mem.h"
#include "csb437_gpio.h"
#include "cpuio.h"

extern  void putpsr(ulong);
extern  ulong getpsr();

extern ulong i2c_init();
extern void ps2_kbd_init(void);
extern ulong ps2_gotachar(void);
extern uchar ps2_getchar(void); 
extern int GPIO_out(int);
extern int GPIO_in(int);
extern void GPIO_init(void);
extern int GPIO_tst(int);
extern void GPIO_set(int);
extern void GPIO_clr(int);
extern void lcd_putchar(char c);
extern void lcd_init(void);
extern void lcd_on(void);

// Delay for some usecs. - Not accurate, assumes ROM mode
// and no Cache.
void udelay(int delay)
{
  int i;
  for ( i = LOOPS_PER_USEC * delay; i ; i--)
	;
}

void rtc_init(void)
{
	// copy DS1307 time/date into cpu
	// not used right now
}

/* InitUART():
 * This is the top-level UART initialization for the monitor. 
 */
void
InitUART(int baud)
{

    if (baud == 0)
        baud = ConsoleBaudRate;

    /* Keep track of the last baud rate, so that it can be used if the */
    /* incoming baudrate is NULL. */
    ConsoleBaudRate = baud;

	// clear error bits and reset tx and rx
	DBGU_REG(DBGU_CR) = DBGU_CR_RSTSTA
					   | DBGU_CR_RSTTX
					   | DBGU_CR_RSTRX;

	// clear any pending interrupts and mask all further interrupts
	DBGU_REG(DBGU_IDR) = DBGU_INT_ALL;
	DBGU_REG(DBGU_IMR) = 0x00000000;

	// Baud divisor = MCK / 16 / baud
	DBGU_REG(DBGU_BRGR) = ((MCK/16)/baud);	// integer only

	// set the port to no parity, no loopback
	DBGU_REG(DBGU_MR) = DBGU_MR_PAR_NONE | DBGU_MR_CHMODE_NORM;

	// Enable
	DBGU_REG(DBGU_CR) = DBGU_CR_TXEN | DBGU_CR_RXEN;

    InitRemoteIO();
}

void
cf_init(void)
{
	// read the CD signal to see if we have a card
	if (GPIO_tst(CF_CD)){
#ifdef CF_DEBUG
		printf("No Compact FLASH Card Detected.\n");
#endif
		return;
	}

#ifdef CF_DEBUG
	printf("Compact FLASH Card Detected.\n");
#endif

	// assign cs4 to CF
	EBI_REG(EBI_CSA) |= EBI_CSA_CS4_CF;

	// make sure the SMC controls CF_CE1/2 and we enable WAIT
	PIOC_REG(PIO_PDR) = A_CF_CE1 | A_CF_CE2 | A_WAIT;

	// set CS4 for 10 wait states, 16-bits
	SMC_REG(SMC_CSR4) = SMC_CSR_RWHOLD(2)
					  | SMC_CSR_RWSETUP(4)
					  | SMC_CSR_ACSS_3
					  | SMC_CSR_DRP_NORMAL
					  | SMC_CSR_DBW_16
					  | SMC_CSR_BAT_16_2
					  | SMC_CSR_TDF(4)
					  | SMC_CSR_WSEN
					  | SMC_CSR_NWS(20);

	// enable power to the card
	GPIO_out(CF_RST );	// make it an output
	GPIO_set(CF_RST);	// hold card in reset

	GPIO_out(CF_PWR);	// make it an output
	GPIO_set(CF_PWR);	// turn card power on
	udelay (100);

	GPIO_clr(CF_RST);	// release card reset
	udelay (10000);

}

int
devInit(int baud)
{
	// set CS2 to 3 waits, 16-bits, 1 clock
	// address and strobe delay
	SMC_REG(SMC_CSR2) = 0x11003183;

	// set CS3 for 10 wait states, 8-bits
	SMC_REG(SMC_CSR3) = 0x1100418a;

	// Initialize the UART
    InitUART(baud);

	// Initialize the GPIO
//	GPIO_init();

	// Turn on the alert LED
	GPIO_out(USR_LED);	// make it an output
	GPIO_clr(USR_LED);	// turn it on

	// Initialize the I2C Port
	i2c_init();

	// Initialize the CF Socket
	cf_init();

	// Initialize the SM501 Video Controller
//	smi_init();
//	smi_on();

	// Initialize the PS2 controller and keyboard - if attached
//	ps2_kbd_init();

	return 0;
}

int
ConsoleBaudSet(int baud)
{
	// Baud divisor = MCK / 16 / baud
	DBGU_REG(DBGU_BRGR) = ((MCK/16)/baud);	// integer only
	return(0);
}

/* target_console_empty():
 * Target-specific portion of flush_console() in chario.c.
 * This function returns 1 if there are no characters waiting to
 * be put out on the UART; else return 0 indicating that the UART
 * is still busy outputting characters from its FIFO.
 */
int
target_console_empty(void)
{
	if ((DBGU_REG(DBGU_CSR) & DBGU_INT_TXBUFE))
		return(1);
	return(0);
}

/* target_putchar():
 */
int
target_putchar(char c)
{
    /* Wait for transmit ready bit
	 */
    while(1) {
        if (DBGU_REG(DBGU_CSR) & DBGU_INT_TXRDY)
            break;
    }
	DBGU_REG(DBGU_THR) = c;
    return((int)c);
}

/* target_gotachar():
 * Return 0 if no char is avaialable at UART rcv fifo; else 1.
 * Do NOT pull character out of fifo, just return status. 
 */
int
target_gotachar()
{
    if (!(DBGU_REG(DBGU_CSR) & DBGU_INT_RXRDY))
        return(0);
    return(1);
}

/* target_getchar():
 */
int 
target_getchar(void)
{
    char    c;

    /* Retrieve character */
    c = DBGU_REG(DBGU_RHR);
    return((int)c);
}

void
initCPUio()
{
}

void
intsrestore(psr)
ulong     psr;
{
    putpsr(psr);
}

/*
 * Read the program status register (CPSR)
 * and set the FIQ and IRQ bits.
 */
ulong
intsoff()
{
    ulong  psr;

    psr = getpsr();

    /*
     * Set bit 6, bit 7 to disable interrupts.
     */
    putpsr(psr | 0x000000c0);
    return(psr);
}

/* cacheInitForTarget():
    Enable instruction cache only...
*/
void
cacheInitForTarget()
{
    asm("   MRC p15, 0, r0, c1, c0, 0");
    asm("   ORR r0, r0, #0x1000");  /* bit 12 is ICACHE enable*/
    asm("   MCR p15, 0, r0, c1, c0, 0");

    /* Flush instruction cache */
    asm("   MCR p15, 0, r0, c7, c5, 0");
}

void
target_reset(void)
{
	coldstart();
}
