//==========================================================================
//
// mc9328mxl_lcd.h
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         11/04/2004
// Description:  This file contains register offsets and bit defines
//				 for the MC9328MXL LCD Controller
//

#include "bits.h"

#define LCDC_SSA 				0x00 		// Screen Start Address Register
#define LCDC_SIZE 				0x04 		// Size Register
#define LCDC_VPW 				0x08 		// Virtual Page Width Register
#define LCDC_CPOS 				0x0C 		// Cursor Position Register
#define LCDC_CWHB 				0x10 		// Cursor Width Height and Blink Register
#define LCDC_CCMR 				0x14 		// Color Cursor Mapping Register
#define LCDC_PCR 				0x18 		// Panel Configuration Register
#define LCDC_HCR 				0x1C 		// Horizontal Configuration Register
#define LCDC_VCR 				0x20 		// Vertical Configuration Register
#define LCDC_POS 				0x24 		// Panning Offset Register
#define LCDC_GPMR 				0x28 		// Gray Palette Mapping and Sharp HR-TFT Configuration Register
#define LCDC_PWMR 				0x2C 		// PWM Contrast Control Register
#define LCDC_DMACR 				0x30 		// DMA Control Register
#define LCDC_RMCR 				0x34 		// Refresh Mode Control Register
#define LCDC_ICR 				0x38 		// Interrupt Configuration Register
#define LCDC_ISR 				0x40 		// Interrupt Status Register

// Bit defines - Refer to MC9328MXL Users Manual Section 17 for more detailed information
//
//  LCDC_SSA - Screen Start Address Register
#define LCDC_SSA_MASK			0xfffffffc	// must be on a 4byte boundry

//  LCDC_SIZE - Size Register
#define LCDC_SIZE_XMAX(_x_)		((_x_ & 0x1ff) << 20)	// X  = (width of panel in pixels)/16, for mono = (width/32)
#define LCDC_SIZE_YMAX(_x_)		((_x_ & 0xfff) << 0)	// Y  = height of panel in lines

//  LCDC_VPW - Virtual Page Width Register
#define LCDC_VPW_MASK			0x3ff

//  LCDC_CPOS - Cursor Position Register
#define LCDC_CPOS_CC_TRANS		(0 << 30)	// OP = 0: cursor is disabled
#define LCDC_CPOS_CC_BW			(1 << 30)	// OP = 0: Full black/color cursor
#define LCDC_CPOS_CC_REV		(2 << 30)	// OP = 0: Reversed video
#define LCDC_CPOS_CC_WHITE		(3 << 30)	// OP = 0: Full white cursor
#define LCDC_CPOS_CC_NO_CURS 	(0 << 30)	// OP = 1: cursor is disabled
#define LCDC_CPOS_CC_OR_CURS 	(1 << 30)	// OP = 1: OR background and cursor
#define LCDC_CPOS_CC_XOR_CURS	(2 << 30)	// OP = 1: XOR background and cursor
#define LCDC_CPOS_CC_AND_CURS	(3 << 30)	// OP = 1: AND background and cursor
#define LCDC_CPOS_CC_OP_EN		BIT28		// 1 = Enable cursor arithmetic operation
#define LCDC_CPOS_CC_CXP(_x_)	((_x_ & 0x3ff) << 16)	// cursor X starting position in pixels
#define LCDC_CPOS_CC_CYP(_x_)	((_x_ & 0x1ff) << 0)	// cursor Y starting position in lines

//  LCDC_CWHB - Cursor Width Height and Blink Register
#define LCDC_CWHB_BK_EN			BIT31		// 1 = Blink is enabled
#define LCDC_CWHB_CW(_x_)		((_x_ & 0x1f) << 24)	// width of the hardware cursor in pixels.
#define LCDC_CWHB_CH(_x_)		((_x_ & 0x1f) << 16)	// height of the hardware cursor in pixels ( 0 disables cursor)
#define LCDC_CWHB_BD(_x_)		((_x_ & 0xff) << 0)		// blink rate = 32Khz divded by this value

//  LCDC_CCMR - Color Cursor Mapping Register
#define LCDC_CCMR_CUR_COL_R(_x_)	((_x_ & 0x1f << 11)	// Red component of cursor
#define LCDC_CCMR_CUR_COL_G(_x_)	((_x_ & 0x3f << 5)	// Green component of cursor
#define LCDC_CCMR_CUR_COL_B(_x_)	((_x_ & 0x1f << 0)	// Blue component of cursor

//  LCDC_PCR - Panel Configuration Register
#define LCDC_PCR_TFT			BIT31		// 1 = The LCD panel is an active display - forces bus width = 16
#define LCDC_PCR_PASSIVE		0
#define LCDC_PCR_COLOR			BIT30		// 1 = The LCD panel is a color display (must set to 1 if TFT is 1)
#define LCDC_PCR_MONO			0
#define LCDC_PCR_PBSIZ1			(0 << 28)	// 1-bit panel bus width - Mono Passive Only
#define LCDC_PCR_PBSIZ2			(1 << 28)	// 2-bit panel bus width - Mono Passive Only
#define LCDC_PCR_PBSIZ4			(2 << 28)	// 4-bit panel bus width - Mono Passive Only
#define LCDC_PCR_PBSIZ8			(3 << 28)	// 8-bit panel bus width - Color Passive Only
#define LCDC_PCR_BPIX1			(0 << 25)	// 1 bpp, FRC bypassed
#define LCDC_PCR_BPIX2			(1 << 25)	// 2 bpp 
#define LCDC_PCR_BPIX4			(2 << 25)	// 4 bpp
#define LCDC_PCR_BPIX8			(3 << 25)	// 8 bpp
#define LCDC_PCR_BPIX12			(4 << 25)	// 12 bpp when TFT = 0
#define LCDC_PCR_BPIX16			(4 << 25)	// 16 bpp when TFT = 1
#define LCDC_PCR_PIXPOL_LO		BIT24		// Pixel Polarity, 0 = Active high, 1 = Active low
#define LCDC_PCR_PIXPOL_HI		0
#define LCDC_PCR_FLMPOL_LO		BIT23		// FLM Polarity, 0 = Active high, 1 = Active low
#define LCDC_PCR_FLMPOL_HI		0
#define LCDC_PCR_LPPOL_LO		BIT22		// LP Polarity, 0 = Active high, 1 = Active low
#define LCDC_PCR_LPPOL_HI		0
#define LCDC_PCR_CLKPOL_TFT_HI	BIT21		// LSCLK Polarity, 0 = falling edge, 1 = rising edge (inverted for TFT)
#define LCDC_PCR_CLKPOL_TFT_LO	0
#define LCDC_PCR_CLKPOL_PAS_LO	BIT21		// LSCLK Polarity, 0 = falling edge, 1 = rising edge (inverted for TFT)
#define LCDC_PCR_CLKPOL_PAS_HI	0
#define LCDC_PCR_OEPOL_LO		BIT20		// OE Polarity, 0 = Active high, 1 = Active low
#define LCDC_PCR_OEPOL_HI		0
#define LCDC_PCR_SCLKIDLE		BIT19		// 0 = Disable LSCLK during VSYNC, 1 = Enable LSCLK
#define LCDC_PCR_END_BIG		BIT18		// 0 = Little endian image format, 1 = Big endian
#define LCDC_PCR_END_LITTLE		0
#define LCDC_PCR_SWAP_DATA		BIT17		// Data Swap for Little endian Mode, 1 = 8 bpp. 4 bpp, 2 bpp, 1 bpp mode
#define LCDC_PCR_SWAP_REV_VS	BIT16		// 1 = Vertical scan in reverse direction
#define LCDC_PCR_SWAP_ACDSEL	BIT15		// 0 = Clock ACD by FRM, 1 = Toggle LP/HSYN
#define LCDC_PCR_SWAP_ACD(_x_)	((_x_ & 0x7f) << 8)	// ACBias divider from clock (based on ACDSEL, ignored for TFT)
#define LCDC_PCR_SWAP_SCLKSEL	BIT7		// 1 = Always enable LSCLK in TFT mode even if there is no data output
#define LCDC_PCR_SWAP_SHARP		BIT6		// 1 = Enable Sharp signals - TFT mode only
#define LCDC_PCR_SWAP_PCD(_x_)	((_x_ & 0x3f) << 0)	// Pixel Clock Divisor (N + 1) - must be 1/5th or less of HCLK

//  LCDC_HCR - Horizontal Configuration Register
#define LCDC_HCR_H_WIDTH(_x_)	((_x_ & 0x3f) << 26)	// HSYNC Pulse Width in pixels, (N + 1)
#define LCDC_HCR_H_WAIT_1(_x_)	((_x_ & 0xff) << 8)		// OE to HSYNC (front porch), (N + 1) 
#define LCDC_HCR_H_WAIT_2(_x_)	((_x_ & 0xff) << 0)		// HSYNC to OE (back porch), (N + 3) 

//  LCDC_VCR - Vertical Configuration Register
#define LCDC_VCR_V_WIDTH(_x_)	((_x_ & 0x3f) << 26)	// VSYNC Pulse Width in lines
#define LCDC_VCR_V_WAIT_1(_x_)	((_x_ & 0xff) << 8)		// OE to VSYNC (front porch) 
#define LCDC_VCR_V_WAIT_2(_x_)	((_x_ & 0xff) << 0)		// VSYNC to OE (back porch) 

//  LCDC_GPMR - Gray Palette Mapping Register
#define LCDC_GPMR_PS_RISE_DELAY(_x_)	((_x_ & 0x3f) << 26)	// CLS to PS delay in PerCLK2's
#define LCDC_GPMR_CLS_RISE_DELAY(_x_)	((_x_ & 0xff) << 16)	// HSYNC to CLS delay in PerCLK2's (N + 1)
#define LCDC_GPMR_REV_TOGGLE_DELAY(_x_)	((_x_ & 0x0f) << 8)		// HSYNC to REV delay in PerCLK2's (N + 1)
#define LCDC_GPMR_GRAY2(_x_)			((_x_ & 0x0f) << 4)		// Gray Scale intermediate value 2
#define LCDC_GPMR_GRAY1(_x_)			((_x_ & 0x0f) << 0)		// Gray Scale intermediate value 1

//  LCDC_PWMR - PWM Contrast  Control Register
#define LCDC_PWMR_CLS_HI_WIDTH(_x_)	((_x_ & 0x1ff) << 16)	// CLS Width Hi in SCLK's (N + 1)
#define LCDC_PWMR_LDMSK				BIT15					// 1 = LD0-15 = 0 during power off
#define LCDC_PWMR_SCR_LINE			0 << 9					// PWM Source Clock = Line Pulses
#define LCDC_PWMR_SCR_PIXEL			1 << 9					// PWM Source Clock = Pixel Clock
#define LCDC_PWMR_SCR_LCD			2 << 9					// PWM Source Clock = LCD Block Clock
#define LCDC_PWMR_CC_EN				BIT8					// 1 = Contrast control is on
#define LCDC_PWMR_PW(_x_)			((_x_ & 0xff) << 0)		// PWM Pulse Width

//  LCDC_DMACR - DMA Control Register
#define LCDC_DMACR_BURST_FIXED		BIT31		// 0 = DMA Burst length is dynamic, 1 = fixed
#define LCDC_DMACR_BURST_DYNAMIC	0
#define LCDC_DMACR_HM(_x_)			((_x_ & 0x0f) << 16	// DMA High Water Mark
#define LCDC_DMACR_TM(_x_)			((_x_ & 0x0f) << 0	// DMA Low Water Mark

//  LCDC_RMCR - Refresh Mode Control Register
#define LCDC_RMCR_LCDC_ON		BIT1		// 1 = Enable the LCDC
#define LCDC_RMCR_SELF_REF		BIT0		// 1 = Enable self-refresh

//  LCDC_ICR - Interrupt Configuration Register
#define LCDC_ICR_INTSYN			BIT2		// 0 = set interrupt on loading data, 1 = outputting data
#define LCDC_ICR_INTCON			BIT0		// 0 = Interrupt on End of Frame, 1 = Start of Frame

//  LCDC_ISR - Interrupt Status Register
#define LCDC_ISR_UDR_ERR		BIT3		// 1 = Underrun Error
#define LCDC_ISR_ERR_RES		BIT2		// 1 = Error from Memory to LCD DMA Request
#define LCDC_ISR_EOF			BIT1		// 1 = End Of Frame has occured
#define LCDC_ISR_SOF			BIT0		// 1 = Start Of Frame has occured



