//=============================================================================
//
//      etherdev.c
//
//      Ethernet Abstraction Layer for Micro Monitor
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         05-26-2002
// Description:  This file contains the interface layer between Micro Monitor
//               and the Ethernet driver for the cs8900a on the CSB226.
//
// Modified by Technology Rescue.
//	25.11.2003 : ddlc added IPADD and associated vars init from DS1307 RAM.
//	07.05.2004 : ddlc added cs8900a_get_mac in extGetEtherAdd function.
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "stddefs.h"
#include "ether.h"
#include "cs8900a.h"
#include "ds1307.h"
#include "cli.h"
#include "i2c.h"


ulong tx_buffer[450];
uchar *tx_buf;

#if INCLUDE_ETHERNET

char IpAddress[18];	/* Local var to store IP address in extGetIpAdd func. */
char *EnvNames[MAX_IDX] =
	{	IP_ADDR_NAME,
		NETMASK_NAME,
		GIPADD_NAME,
		DNS1_NAME,
		DNS2_NAME,
		HOSTIP_NAME,
		ETHERADD_NAME,
	};

extern unsigned char BinEnetAddr[6];						// Space for binary MAC address.
char MacAddress[18];


/*
 * enreset():
 *	Reset the PHY and MAC.
 */
void
enreset(void)
{
	cs8900a_reset();
}

/*
 * eninit():
 * This would include establishing buffer descriptor tables and
 * all the support code that will be used by the ethernet device.
 *
 * It can be assumed at this point that the array uchar BinEnetAddr[6]
 * contains the 6-byte MAC address.
 *
 * Return 0 if successful; else -1.
 */
int
eninit(void)
{
	/* align tx_buf on a 64-byte boundary:
	 */
	tx_buf = (uchar *)tx_buffer;
	while((ulong)tx_buf & 0x3f)
		tx_buf++;

	return cs8900a_init();
}

int
EtherdevStartup(int verbose)
{
	/* Initialize local device error counts (if any) here. */
	/* OPT_ADD_CODE_HERE */

	/* Put ethernet controller in reset: */
	enreset();

	/* Initialize controller: */
	eninit();

	return(0);
}

/* disablePromiscuousReception():
 * Provide the code that disables promiscuous reception.
 */
void
disablePromiscuousReception(void)
{
	/* OPT_ADD_CODE_HERE */	 
}

/* enablePromiscuousReception():
 * Provide the code that enables promiscuous reception.
 */
void
enablePromiscuousReception(void)
{
	/* OPT_ADD_CODE_HERE */
}

/* disableBroadcastReception():
 * Provide the code that disables broadcast reception.
 */
void
disableBroadcastReception(void)
{
	/* ADD_CODE_HERE */
}

/* enableBroadcastReception():
 * Provide the code that enables broadcast reception.
 */
void
enableBroadcastReception(void)
{
	/* ADD_CODE_HERE */
}

/* 
 * enselftest():
 *	Run a self test of the ethernet device(s).  This can be stubbed
 *	with a return(1).
 *	Return 1 if success; else -1 if failure.
 */
int
enselftest(int verbose)
{
	return(1);
}

/* ShowEtherdevStats():
 * This function is used to display device-specific stats (error counts
 * usually).
 */
void
ShowEtherdevStats(void)
{
	/* OPT_ADD_CODE_HERE */
}

/* getXmitBuffer():
 * Return a pointer to the buffer that is to be used for transmission of
 * the next packet.  Since the monitor's driver is EXTREMELY basic,
 * there will only be one packet ever being transmitted.  No need to queue
 * up transmit packets.
 */
uchar *
getXmitBuffer(void)
{
	return(tx_buf);
}

/* sendBuffer():
 * Send out the packet assumed to be built in the buffer returned by the
 * previous call to getXmitBuffer() above.
 */
int
sendBuffer(int length)
{
	int temp32;

    if (length < 64)
        length = 64;

#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_OUTGOING)
		printPkt((struct ether_header *)tx_buf,length,ETHER_OUTGOING);
#endif

	// tell the cs8900a to send the tx buffer pointed to by tx_buf
	temp32 = cs8900a_tx(tx_buf, length);

	EtherXFRAMECnt++;
	if (temp32) {
		return -1;
	}
	else {
		return 0;
	}
}

/* DisableEtherdev():
 * Fine as it is...
 */
void
DisableEtherdev(void)
{
	enreset();
}

/* extGetIpAdd():
 * If there was some external mechanism (other than just using the
 * IPADD shell variable established in the monrc file) for retrieval of
 * the board's IP address, then do it here...
 */
char *
extGetIpAdd(void)
{
#if 0	

	char *RetVal = DEFAULT_IPADD;					// To return from.
	int error, Index;
	ulong IpAddr;									// To store IP addr read from RTC RAM.
	int IpAddrIsFixed = 0;							// Assume IPADD read from RTC RAM is not fixed.
	char TempAddress[18];							// Temporary buffer to set shell vars.

	/* Read IP address from RTC I2C RAM. */
	error = i2cRead(DS1307, IP_ADDR_OFFSET, (char *)(&IpAddr), IP_ADDR_SIZE);
	if(error<0)
	{
		#ifdef I2C_DEBUG
		printf("i2cRead error: %d\n", error);
		#endif
	}
	else
	{
		/* Check if IP address is valid. */
		if(IpAddr != 0xFFFFFFFF && IpAddr != 0x00)
		{	// If IP Address is valid.

			if(IpAddr == BOOTP_ID)
			{	strcpy(IpAddress, "BOOTP");
			}
			else if(IpAddr == DHCP_ID)
			{	strcpy(IpAddress, "DHCP");
			}
			else
			{
				IpToString(IpAddr, IpAddress);
				setenv(EnvNames[IP_ADDR_IDX],IpAddress);
				IpAddrIsFixed = 1;					// In this case, IPADD is fixed.
			}
			// IpAddress is used to return address from memory, not from stack.
			#ifdef I2C_DEBUG
			printf("IP ADD = %x %s\n", IpAddr, IpAddress);
			#endif
			RetVal = IpAddress;
		}
	}

	/* Create remaining shell variables here exept HOSTIP, if IPADD is fixed. */
	if(IpAddrIsFixed)
	{
		for(Index = NETMASK_IDX; Index < HOSTIP_IDX; Index++)
		{
			/* Read address from I2C RTC RAM.
			   Start at address IP_ADDR_OFFSET + 4 * 1.
			*/
			error = i2cRead(DS1307, IP_ADDR_OFFSET + (IP_SIZE * Index), (char *)(&IpAddr), IP_SIZE);
			if(error<0)
			{
				#ifdef I2C_DEBUG
				printf("i2cRead error: %d\n", error);
				#endif
			}
			else
			{	if(IpAddr != 0xFFFFFFFF && IpAddr != 0x00)
				{
					// convert netmask to shell var.
					IpToString(IpAddr, TempAddress);
					setenv(EnvNames[Index],TempAddress);
				}
			}
		}
	}

	/* Read HOSTIP from RTC RAM. */
	error = i2cRead(DS1307, HOSTIP_OFFSET, (char *)(&IpAddr), HOSTIP_SIZE);
	if(error<0)
	{
		#ifdef I2C_DEBUG
		printf("i2cRead error: %d\n", error);
		#endif
	}
	else
	{	if(IpAddr != 0xFFFFFFFF && IpAddr != 0x00)
		{
			// convert gateway address to shell var.
			IpToString(IpAddr, TempAddress);
			setenv(EnvNames[HOSTIP_IDX],TempAddress);
		}
	}

	return RetVal;
#else
	return((char *)0);
#endif
}

/* extGetEtherAdd():
 * If there was some external mechanism (other than just using the
 * ETHERADD shell variable established in the monrc file) for retrieval of
 * the board's MAC address, then do it here...
 */
char *
extGetEtherAdd(void)
{
#if 0
	char *RetVal = DEFAULT_ETHERADD;				// To return from.
	int error = 0;

	if(cs8900a_get_mac() == 0)
	{	/* MAC address from CS8900 registers are not valid. */

		/* Read MAC address from RTC I2C RAM. */
		error = i2cRead(DS1307, ETHERADD_OFFSET, (char *)(&BinEnetAddr), ETHERADD_SIZE);
		if(error<0)
		{
			#ifdef I2C_DEBUG
			printf("i2cRead error: %d\n", error);
			#endif
		}
	}
	if(error >=0)
	{
		EtherToString(BinEnetAddr, MacAddress);
		setenv(EnvNames[ETHERADD_IDX],MacAddress);
		RetVal = MacAddress;
	}

	return RetVal;
#else
	return((char *)0);
#endif

}

/*
 * polletherdev():
 * Called continuously by the monitor (ethernet.c) to determine if there
 * is any incoming ethernet packets.
 */
int
polletherdev(void)
{
	static ulong pktbuf[RBUFSIZE/4];
	int	pktlen = 0, pktcnt = 0;

//	while(1)
//	{
		pktlen = cs8900a_rx((uchar *)pktbuf);
//		if (pktlen == 0) break;

	if(pktlen){
		pktcnt++;
		EtherRFRAMECnt++;
		processPACKET((struct ether_header *)pktbuf, pktlen);
	}
	return(pktcnt);
}

char *write_paramHelp[] = {
	"This command saves parameters into I2C RTC RAM.",
    "[{IP add} or BOOTP or DHCP]",
    "                   [{netmask}]",
    "                   [{Gateway IP add}]",
    "                   [{Primary DNS}]",
    "                   [{Secondary DNS}]",
    "                   [{Host IP address}]",
	"                   [{MAC address}]",
    "Writes IP address, sub-netmask, Gateway IP and DNS addesses to I2C RTC RAM.",
	"IP address is required.",
    "DNS are optional."
	"To remove DNS, write '255.255.255.255'",
	"To keep parameter unchanged, write '0'.",
	"Host IP address parameter is used to connect WindowsCE kernel to Host.",
	"MAC address is optional. It should be set once.",
    0
};

int write_param(int argc,char *argv[])
{
	ulong error, Index;
	uchar IpAddConv[IP_SIZE];				// To store converted IP addr.
	uchar EtherAddConv[ETHERADD_SIZE];		// To store converted Ether Addr.

    if (argc <2 ||argc > 8)
        return(CMD_PARAM_ERROR);

	/* argv[1] : new IP address to write to RTC RAM. */
	if(!strcmp(argv[1], "BOOTP"))
	{	// argv[1] = BOOTP
		memcpy((char *)IpAddConv, "BOTP", IP_SIZE);
	}
	else if(!strcmp(argv[1], "DHCP"))
	{	// argv[1] = DHCP
		memcpy((char *)IpAddConv, "DHCP", IP_SIZE);
	}
	else
	{
		if(IpToBin(argv[1], IpAddConv))		// Convert arg to binary form.
			return (CMD_FAILURE);					
	}

	/* update I2C RTC RAM. */
	error = i2cWrite(DS1307, IP_ADDR_OFFSET, IpAddConv, IP_ADDR_SIZE);
	if(error<0)
	{
		#ifdef I2C_DEBUG
		printf("write_param error: %d\n", error);
		#endif
		return (CMD_FAILURE);
	}
	/* Update shell variable if write to RTC RAM succeeded. */
	setenv(EnvNames[IP_ADDR_IDX],argv[1]);

	/* argv[2] : new NETMASK to write to RAM. */
	/* argv[3] : new GIPADD to write to RAM. */
	/* argv[4] : new primary DNS to write to RAM. */
	/* argv[5] : new secondary DNS to write to RAM. */
	/* argv[6] : new HOST IP address to write to RAM. */

	for(Index = NETMASK_IDX; Index <= HOSTIP_IDX; Index++)
	if(argc > Index+1)
	{
		if(strcmp(argv[Index+1],"0") && !IpToBin(argv[Index+1], IpAddConv))
		{
			/* update I2C RTC RAM.
			   Start at address IP_ADDR_OFFSET + 4 * 1.
			*/
			error = i2cWrite(DS1307, IP_ADDR_OFFSET + (IP_SIZE * Index), IpAddConv, IP_SIZE);
			if(error<0)
			{
				#ifdef I2C_DEBUG
				printf("write_param error: %d\n", error);
				#endif
				return (CMD_FAILURE);
			}
			/* Update shell variable if write to RTC RAM succeeded. */
			setenv(EnvNames[Index],argv[Index+1]);
		}
	}

	/* argv[7] : new MAC address to write to RAM. */
	if(argc > Index+1)
	{
		if(strcmp(argv[Index+1],"0") && !EtherToBin(argv[Index+1], EtherAddConv))
		{
			/* update I2C RTC RAM. */
			error = i2cWrite(DS1307, ETHERADD_OFFSET, EtherAddConv, ETHERADD_SIZE);
			if(error<0)
			{
				#ifdef I2C_DEBUG
				printf("write_param error: %d\n", error);
				#endif
				return (CMD_FAILURE);
			}
			/* Update shell variable if write to RTC RAM succeeded. */
			setenv(EnvNames[ETHERADD_IDX],argv[ETHERADD_IDX+1]);
		}
	}

	return(CMD_SUCCESS);
}

char *read_paramHelp[] = {
	"This command verifies parameters from I2C RTC RAM.",
    "",
    "", 
    "", 
    0
};

int read_param(int argc,char *argv[])
{
	int Index;

    if (argc != 1)
        return(CMD_PARAM_ERROR);

	strcpy(IpAddress,extGetIpAdd());					// Update IpAddress var.
	printf("IP address = %s\n", IpAddress);
	if(strcmp(IpAddress, "BOOTP") && strcmp(IpAddress, "DHCP") )
	{
		for(Index = NETMASK_IDX; Index < HOSTIP_IDX; Index++)
		printf("%s = %s\n", EnvNames[Index], getenv(EnvNames[Index]));
	}
	printf("%s = %s\n", EnvNames[HOSTIP_IDX], getenv(EnvNames[HOSTIP_IDX]));
	printf("%s = %s\n", EnvNames[ETHERADD_IDX], getenv(EnvNames[ETHERADD_IDX]));
	return(CMD_SUCCESS);
}

#endif
