//==========================================================================
//
//      cpu_i2c.c
//
// Author(s):   Michael Kelly - Cogent Computer Systems, Inc.
// Date:        03/26/2003
// Description:	Generic IIC Routines - requires I2C_SCL and I2C_SDA to
//				be defined in cpu_gpio.h
//
//==========================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "stddefs.h"
#include "cli.h"
#include "cpu_gpio.h"
#include "cpu_i2c.h"
#include "umongpio.h"

//--------------------------------------------------------------------------
// i2c_init()
//
// Initialize the I2C registers for direct I2C control
ulong i2c_init()
{
	I2C_REG(I2CR) = 0;	// Reset module
	I2C_REG(IFDR) = 0x38;	// ipg_clk (~33MHz) / 640 = ~50Khz 
	I2C_REG(I2CR) = I2CR_IEN;	// 
	I2C_REG(I2SR) = 0;	// 

	return 0;
}

//--------------------------------------------------------------------------
// i2c_wr_device()
//
// This function writes an 8-bit value to the I2C device at the requested
// register.
//
ulong i2c_wr_device(uchar dev, uchar reg, uchar data)
{

	// clear any left over interrupts
	I2C_REG(I2SR) = 0; 
	
	// write the 7-bit device address with write = 0
	// this write needs a start command (created by a
	// 0 to 1 transition of the MSTA bit)
	I2C_REG(I2CR) = I2CR_IEN | I2CR_MSTA | I2CR_MTX;
	
	if(i2c_wr_byte((dev << 1) & 0xfe)){
		printf("I2C Bus Write dev Failure - Check Paramters!\n");
		return -1;
	}
	// Write the 8-bit register address
	if(i2c_wr_byte(reg)){
		printf("I2C Bus Write reg Failure - Check Paramters!\n");
		return -1;
	}
	// Write the 8-bit data value
	if(i2c_wr_byte(data)){
		printf("I2C Bus Write data Failure - Check Paramters!\n");
		return -1;
	}

	// issue a stop (created by a 1 to 0 transition
	// of the MSTA bit)
	I2C_REG(I2CR) = I2CR_IEN;
	
	return 0;
}

//--------------------------------------------------------------------------
// i2c_rd_device()
//
// This function reads an 8-bit value from the I2C device at the requested
// register.
//
ulong i2c_rd_device(uchar dev, uchar reg, uchar *data)
{

	//I2C_REG(I2CR) = 0;
	//I2C_REG(I2CR) = I2CR_IEN;
	//udelay(10);

	// clear any left over interrupts
	I2C_REG(I2SR) = 0; 

	// write the 7-bit device address with write = 0
	// this write needs a start command (created by a
	// 0 to 1 transition of the MSTA bit)
	I2C_REG(I2CR) = I2CR_IEN | I2CR_MSTA | I2CR_MTX;
	
	if(i2c_wr_byte((dev << 1) & 0xfe)){
		printf("I2C Bus Write dev Failure - Check Paramters!\n");
		return -1;
	}

	// Write the 8-bit register address
	if(i2c_wr_byte(reg)){
		printf("I2C Bus Write reg Failure - Check Paramters!\n");
		return -1;
	}

	// write the 7-bit device address again with write = 1
	// this write needs a repeat start command (created by a
	// 0 to 1 transition of the RSTA bit)
	I2C_REG(I2CR) = I2CR_IEN | I2CR_MSTA | I2CR_MTX | I2CR_RSTA;
	
	if(i2c_wr_byte((dev << 1) | 1)){
		printf("I2C Bus Write restart Failure - Check Paramters!\n");
		return -1;
	}

	// Transmit ack needs to be set
	I2C_REG(I2CR) = I2CR_IEN | I2CR_MSTA | I2CR_TX_NO_AK;
	
	// do dummy read of data reg
	*data = I2C_REG(I2DR);

	// read the data, drop MTX to put us into read mode
	*data = i2c_rd_byte();

	// wait for the interrupt flag
	while ((I2C_REG(I2SR) & I2SR_IIF) == 0){}

	// issue a stop (created by a 1 to 0 transition
	// of the MSTA bit)
	I2C_REG(I2CR) = I2CR_IEN;
	
	return 0;
}

//--------------------------------------------------------------------------
// i2c_wr_byte()
//
// This function writes an 8-bit value to the I2C bus, MSB first.
// Data is written by changing SDA during SCL low, then bringing
// SCL high.  SCL is returned low to setup for the next transition.
//
ulong i2c_wr_byte(uchar data)
{

	// write data and wait a bit before checking for ack
	I2C_REG(I2DR) = data;

	// wait for the interrupt flag
	while ((I2C_REG(I2SR) & I2SR_IIF) == 0){}

	if(I2C_REG(I2SR) & I2SR_RX_NO_AK){	// a high means no ack

		I2C_REG(I2SR) = 0;
		return -1;
	}

	I2C_REG(I2SR) = 0;
	return 0;
}

//--------------------------------------------------------------------------
// i2c_rd_byte()
//
// This function reads an 8-bit data value from the I2C bus, MSB first. 
// Data is read from SDA after each low to high SCL transition.
//
uchar i2c_rd_byte()
{
	uchar volatile rd_data;
	rd_data = 0;

	// wait for the completion flag
	while ((I2C_REG(I2SR) & I2SR_IIF) == 0){}
	I2C_REG(I2SR) = 0;
	
	// get the data
	rd_data = I2C_REG(I2DR);
	return rd_data;
}


char *i2cHelp[] = {
	" This command allows the user to read ",
	" and write devices on the i2c bus. \n ",
    " Usage:",
    " i2c -[w] {device} {register} {val/count}",
    " Options...",
    " -w write val to device/register",
	" no options, read from device/register up to count",
	0
};

int i2c(int argc,char *argv[])
{
	int i, opt;
	int write = 0;
	uchar dev, reg, data, count;

	// Initialize the I2C Port
	i2c_init();

	while ((opt=getopt(argc,argv,"w")) != -1) {
    if (opt == 'w') write = 1;
	}

	// make sure we have the right number of paramters
	if (argc < (optind+3))
		return(CMD_PARAM_ERROR);

	dev = (uchar) strtoul(argv[optind],(char **)0,0);
	reg = (uchar) strtoul(argv[optind+1],(char **)0,0);

	// 3rd arg is the data value if it's a write, count if it's a read
	data = (uchar) strtoul(argv[optind+2],(char **)0,0);
	count = data;
	// do it
	if (write)
	{
		printf("Writing 0x%02x to Device 0x%02x @ Register 0x%02x.\n", data, dev, reg);
    	if(i2c_wr_device(dev, reg, data))
    	{
			printf("I2C Bus Failure - Check Paramters!\n");
    		return (CMD_FAILURE);
		}
	}
	else
	{
		for (i = 0; i < count; i++)
    	{
			printf("Read Device 0x%02x, Register 0x%02x = ", dev, reg + i);
			if(i2c_rd_device(dev, reg + i, &data))
			{
				printf("I2C Bus Failure - Check Paramters!\n");
    			return (CMD_FAILURE);
			}
			printf("0x%02x.\n", data);
		}
	}
	return(CMD_SUCCESS);
}

