//==========================================================================
//
// csb737_gpio.c
//
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// 				 Luis Torrico, Cogent Computer Systems, Inc. 
// Date:         12/27/2007
// Description:  This file contains code to intialize the CSB737 GPIO
//               section as well as set, clear and test functions
//               for manipulating the GPIO bits
//
//--------------------------------------------------------------------------

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "umongpio.h"
#include "at91sam9263.h"
#include "at91sam9263_gpio.h"
#include "csb737_gpio.h"

//--------------------------------------------------------
// GPIO_init()
//
// This function sets the startup state for the AT91SAM9263
// GPIO as used on the CSB737.  Refer to csb737_gpio.h for
// a description of the default values.  Here we just put
// them into the chip in the following order:
// 1. Disable all GPIO pins, then set the pins that will be GPIO
// 2. Set the pins that will carry alternate function A
// 3. Set the pins that will carry alternate function B
// 4. Disable all GPIO outputs and disable multi-drive
// 5. Clear the GPIO output pins that need to start low
// 6. Set the GPIO output pins that need to start high
// 7. Disable all pullups
// 8. Set which GPIO pins will have pullups
// 9. Enable the GPIO output pins (I/O pins will be left as inputs)

void GPIO_init()
{
	// Port A
	PIOA_REG(PIO_PDR) = 0xffffffff;						// GPIO Pin Disable
	PIOA_REG(PIO_PER) = ~(DEF_PIOA_ASR | DEF_PIOA_BSR);	// GPIO Pin Enable

	PIOA_REG(PIO_ASR) = DEF_PIOA_ASR;					// Alternate Function A
	PIOA_REG(PIO_BSR) = DEF_PIOA_BSR;					// Alternate Function B

	PIOA_REG(PIO_ODR) = 0xffffffff;						// Disable all GPIO Outputs first
	PIOA_REG(PIO_MDER) = 0x0;							// no multidrive
	PIOA_REG(PIO_SODR) = DEF_PIOB_SODR;					// Set Output Data
	PIOA_REG(PIO_CODR) = DEF_PIOB_CODR;					// Clear Output Data

	PIOA_REG(PIO_PUDR) = 0xffffffff;					// Disable all Pullups first
	PIOA_REG(PIO_PUER) = DEF_PIOA_PUER;					// Enable Desired Pullups 

	PIOA_REG(PIO_OER) = DEF_PIOA_OER;					// GPIO Output Enable
	PIOA_REG(PIO_ODR) = ~(DEF_PIOA_OER);				// GPIO Output Disable (makes the rest inputs)

	// Port B
	PIOB_REG(PIO_PDR) = 0xffffffff;					    // GPIO Pin Disable
	PIOB_REG(PIO_PER) = ~(DEF_PIOB_ASR | DEF_PIOB_BSR); // GPIO Pin Enable

	PIOB_REG(PIO_ASR) = DEF_PIOB_ASR;				    // Alternate Function A
	PIOB_REG(PIO_BSR) = DEF_PIOB_BSR;				    // Alternate Function B

	PIOB_REG(PIO_ODR) = 0xffffffff;					    // Disable all GPIO Outputs first
	PIOB_REG(PIO_MDER) = DEF_PIOB_MDER;					// Enable TWI signals for multidrive
	PIOB_REG(PIO_SODR) = DEF_PIOB_SODR;				    // Set Output Data
	PIOB_REG(PIO_CODR) = DEF_PIOB_CODR;				    // Clear Output Data

	PIOB_REG(PIO_PUDR) = 0xffffffff;				    // Disable all Pullups first
	PIOB_REG(PIO_PUER) = DEF_PIOB_PUER;				    // Enable Desired Pullups 

	PIOB_REG(PIO_OER) = DEF_PIOB_OER;				    // GPIO Output Enable
	PIOB_REG(PIO_ODR) = ~(DEF_PIOB_OER);			    // GPIO Output Disable (makes the rest inputs)

	// Port C
	PIOC_REG(PIO_PDR) = 0xffffffff;						// GPIO Pin Disable
	PIOC_REG(PIO_PER) = ~(DEF_PIOC_ASR | DEF_PIOC_BSR);	// GPIO Pin Enable

	PIOC_REG(PIO_ASR) = DEF_PIOC_ASR;					// Alternate Function A
	PIOC_REG(PIO_BSR) = DEF_PIOC_BSR;					// Alternate Function B

	PIOC_REG(PIO_ODR) = 0xffffffff;						// Disable all GPIO Outputs first
	PIOC_REG(PIO_MDER) = 0x0;							// no multidrive
	PIOC_REG(PIO_SODR) = DEF_PIOC_SODR;					// Set Output Data
	PIOC_REG(PIO_CODR) = DEF_PIOC_CODR;					// Clear Output Data

	PIOC_REG(PIO_PUDR) = 0xffffffff;					// Disable all Pullups first
	PIOC_REG(PIO_PUER) = DEF_PIOC_PUER;					// Enable Desired Pullups 

	PIOC_REG(PIO_OER) = DEF_PIOC_OER;					// GPIO Output Enable
	PIOC_REG(PIO_ODR) = ~(DEF_PIOC_OER);   				// GPIO Output Disable (makes the rest inputs)
	
	// Port D
	PIOD_REG(PIO_PDR) = 0xffffffff;						// GPIO Pin Disable
	PIOD_REG(PIO_PER) = ~(DEF_PIOD_ASR | DEF_PIOD_BSR);	// GPIO Pin Enable

	PIOD_REG(PIO_ASR) = DEF_PIOD_ASR;					// Alternate Function A
	PIOD_REG(PIO_BSR) = DEF_PIOD_BSR;					// Alternate Function B

	PIOD_REG(PIO_ODR) = 0xffffffff;						// Disable all GPIO Outputs first
	PIOD_REG(PIO_MDER) = 0x0;							// no multidrive
	PIOD_REG(PIO_SODR) = DEF_PIOD_SODR;					// Set Output Data
	PIOD_REG(PIO_CODR) = DEF_PIOD_CODR;					// Clear Output Data

	PIOD_REG(PIO_PUDR) = 0xffffffff;					// Disable all Pullups first
	PIOD_REG(PIO_PUER) = DEF_PIOD_PUER;					// Enable Desired Pullups 

	PIOD_REG(PIO_OER) = DEF_PIOD_OER;					// GPIO Output Enable
	PIOD_REG(PIO_ODR) = ~(DEF_PIOD_OER);   				// GPIO Output Disable (makes the rest inputs)

	// Port E
	PIOE_REG(PIO_PDR) = 0xffffffff;						// GPIO Pin Disable
	PIOE_REG(PIO_PER) = ~(DEF_PIOE_ASR | DEF_PIOE_BSR);	// GPIO Pin Enable

	PIOE_REG(PIO_ASR) = DEF_PIOE_ASR;					// Alternate Function A
	PIOE_REG(PIO_BSR) = DEF_PIOE_BSR;					// Alternate Function B

	PIOE_REG(PIO_ODR) = 0xffffffff;						// Disable all GPIO Outputs first
	PIOE_REG(PIO_MDER) = 0x0;							// no multidrive
	PIOE_REG(PIO_SODR) = DEF_PIOE_SODR;					// Set Output Data
	PIOE_REG(PIO_CODR) = DEF_PIOE_CODR;					// Clear Output Data

	PIOE_REG(PIO_PUDR) = 0xffffffff;					// Disable all Pullups first
	PIOE_REG(PIO_PUER) = DEF_PIOE_PUER;					// Enable Desired Pullups 

	PIOE_REG(PIO_OER) = DEF_PIOE_OER;					// GPIO Output Enable
	PIOE_REG(PIO_ODR) = ~(DEF_PIOE_OER);				// GPIO Output Disable (makes the rest inputs)
}
