/*
 * etherdev.c:
 * Ethernet hookup for MCF5329 to uMon.  Slightly modified (by Cogent)
 * from the mcf5329 port for use on CSB760.
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:	Ed Sutter
 * email:	esutter@lucent.com
 * phone:	908-582-2351
 */

#include "config.h"
#include "genlib.h"
#include "stddefs.h"
#include "ether.h"
#include "mcf532x.h"
#include "nbuf.h"
#include "cpuio.h"
#include "cache.h"

#if INCLUDE_ETHERNET

#define MAX_ETH_PKT 1518
#define FULL_DUPLEX
#undef FULL_DUPLEX

#define MCF_FEC_EIR_ALL_EVENTS   \
   ( MCF_FEC_EIR_HBERR | MCF_FEC_EIR_BABR | MCF_FEC_EIR_BABT | \
     MCF_FEC_EIR_GRA   | MCF_FEC_EIR_TXF  | MCF_FEC_EIR_TXB  | \
     MCF_FEC_EIR_RXF   | MCF_FEC_EIR_RXB  | MCF_FEC_EIR_MII  | \
     MCF_FEC_EIR_EBERR | MCF_FEC_EIR_LC   | MCF_FEC_EIR_RL   | \
     MCF_FEC_EIR_UN )

#define MCF_FEC_RxBD_ERRS   \
   ( MCF_FEC_RxBD_LG | MCF_FEC_RxBD_NO |  MCF_FEC_RxBD_CR | \
     MCF_FEC_RxBD_OV | MCF_FEC_RxBD_TR ) 

#if ((RX_BUFFER_SIZE < MAX_ETH_PKT) || (TX_BUFFER_SIZE < MAX_ETH_PKT))
#error "MAX_ETH_PKT is too large."
#endif

static void 
setMII(int phyNumber, int regNumber, int value)
{
	MCF_FEC_MMFR = 
		MCF_FEC_MMFR_ST(1) |
		MCF_FEC_MMFR_OP(1) |
		MCF_FEC_MMFR_PA(phyNumber) |
		MCF_FEC_MMFR_RA(regNumber) |
		MCF_FEC_MMFR_TA(2) |
		MCF_FEC_MMFR_DATA(value);

	while ((MCF_FEC_EIR & MCF_FEC_EIR_MII) == 0) ;
	MCF_FEC_EIR = MCF_FEC_EIR_MII;
}

/*
 * enreset():
 *	Reset the PHY and MAC.
 */
void
enreset(void)
{
	/* Reset the FEC - equivalent to a hard reset */
	MCF_FEC_ECR = MCF_FEC_ECR_RESET;

	/* Wait for the reset sequence to complete */
	while(MCF_FEC_ECR & MCF_FEC_ECR_RESET) ;

	/* Disable all FEC interrupts by clearing the IMR register */
	MCF_FEC_EIMR = 0;

	/* Clear any interrupts by setting all bits in the EIR register */
	MCF_FEC_EIR = MCF_FEC_EIR_ALL_EVENTS;
}

/*
 * eninit():
 * Initialize the PHY and MAC.
 * This would include establishing buffer descriptor tables and
 * all the support code that will be used by the ethernet device.
 *
 * It can be assumed at this point that the array uchar BinEnetAddr[6]
 * contains the 6-byte MAC address.
 *
 * Return 0 if successful; else -1.
 */
int
eninit(void)
{
	/* Disable all FEC interrupts by clearing the IMR register */
	MCF_FEC_EIMR = 0;

	/* Clear any interrupts by setting all bits in the EIR register */
	MCF_FEC_EIR = MCF_FEC_EIR_ALL_EVENTS;

	/* Set the source address for the controller */
	MCF_FEC_PALR = MCF_FEC_PALR_PADDR1(
		(BinEnetAddr[0] <<24) 
		| (BinEnetAddr[1] <<16)	
		| (BinEnetAddr[2] <<8) 
		| (BinEnetAddr[3] <<0));

	MCF_FEC_PAUR = MCF_FEC_PAUR_PADDR2(
		(BinEnetAddr[4] <<8)
		| (BinEnetAddr[5] <<0)) |
		MCF_FEC_PAUR_TYPE(0x8808);

	/* Initialize the hash table registers */
	MCF_FEC_IAUR = 0;
	MCF_FEC_IALR = 0;

	MCF_FEC_GAUR = 0;
	MCF_FEC_GALR = 0;

	/* Set receive buffer size */
	MCF_FEC_EMRBR = MAX_ETH_PKT;

#if defined(FULL_DUPLEX)
	MCF_FEC_RCR = MCF_FEC_RCR_MAX_FL(MAX_ETH_PKT) |
					MCF_FEC_RCR_MII_MODE;
	MCF_FEC_TCR = MCF_FEC_TCR_FDEN;
#else
	MCF_FEC_RCR = MCF_FEC_RCR_MAX_FL(MAX_ETH_PKT) |
					MCF_FEC_RCR_MII_MODE      |
					MCF_FEC_RCR_DRT;
	MCF_FEC_TCR = 0;
#endif

	/* Set MII speed to <= 2.5 MHz */
	MCF_FEC_MSCR = MCF_FEC_MSCR_MII_SPEED(((BUS_FREQUENCY + 5000000 - 1) / 5000000));

	/* Initialize the network buffers */
	nbuf_init();

	/* Point to the start of the circular Rx buffer descriptor queue */
	MCF_FEC_ERDSR = nbuf_get_start(Rx);

	/* Point to the start of the circular Tx buffer descriptor queue */
	MCF_FEC_ETDSR = nbuf_get_start(Tx);

#if defined(FULL_DUPLEX)
	/* Set PHYS to 100 Mb/s, full duplex */
	setMII(31, 0, 0x3100);
#else
	/* Set PHYS to 100 Mb/s, half duplex */
	setMII(31, 0, 0x3080);
#endif

	MCF_FEC_FRSR = MCF_FEC_FRSR_R_FSTART(0x40);


	MCF_FEC_MIBC = MCF_FEC_MIBC_MIB_DISABLE;
	{
		vuint32 *vuip = &MCF_FEC_RMON_T_DROP;
		while (vuip <= &MCF_FEC_IEEE_R_OCTETS_OK)
		*vuip++ = 0;
	}
	MCF_FEC_MIBC = 0;	

	/* Enable FEC */
	MCF_FEC_ECR = MCF_FEC_ECR_ETHER_EN;
	
	/* Indicate that there have been empty receive buffers produced */
	MCF_FEC_RDAR = MCF_FEC_RDAR_R_DES_ACTIVE;

	return (0);
}

int
EtherdevStartup(int verbose)
{
	/* Pin assignments for ports FECH and FECL */
	MCF_GPIO_PDDR_FECH = 0;
	MCF_GPIO_PDDR_FECL = 0;
	MCF_GPIO_PAR_FEC = MCF_GPIO_PAR_FEC_PAR_FEC_7W(0x3) |
						MCF_GPIO_PAR_FEC_PAR_FEC_MII(0x3);

	/* Put ethernet controller in reset: */
	enreset();

	/* Initialize controller and return the value returned by
	 * eninit().
	 */
	return(eninit());
}

/* disablePromiscuousReception():
 * Provide the code that disables promiscuous reception.
 */
void
disablePromiscuousReception(void)
{
	/* OPT_ADD_CODE_HERE */
}

/* enablePromiscuousReception():
 * Provide the code that enables promiscuous reception.
 */
void
enablePromiscuousReception(void)
{
	/* OPT_ADD_CODE_HERE */
}

/* disableBroadcastReception():
 * Provide the code that disables broadcast reception.
 */
void
disableBroadcastReception(void)
{
	/* ADD_CODE_HERE */
}

/* enableBroadcastReception():
 * Provide the code that enables broadcast reception.
 */
void
enableBroadcastReception(void)
{
	/* ADD_CODE_HERE */
}

/* 
 * enselftest():
 *	Run a self test of the ethernet device(s).  This can be stubbed
 *	with a return(1).
 *	Return 1 if success; else -1 if failure.
 */
int
enselftest(int verbose)
{
	return(1);
}

/* ShowEtherdevStats():
 * This function is used to display device-specific stats (error counts
 * usually).
 */
void
ShowEtherdevStats(void)
{
  printf("Rx Packet Count:%-10lu", MCF_FEC_RMON_R_PACKETS);
  printf("   Rx Broadcast:%-10lu\n", MCF_FEC_RMON_R_BC_PKT);
  printf("   Rx Multicast:%-10lu", MCF_FEC_RMON_R_MC_PKT);
  printf("CRC/Align error:%-10lu", MCF_FEC_RMON_R_CRC_ALIGN);
  printf("   Rx Undersize:%-10lu\n", MCF_FEC_RMON_R_UNDERSIZE);
  printf("    Rx Oversize:%-10lu", MCF_FEC_RMON_R_OVERSIZE);
  printf("    Rx Fragment:%-10lu", MCF_FEC_RMON_R_FRAG);
  printf("      Rx Jabber:%-10lu\n", MCF_FEC_RMON_R_JAB);
  printf("          Rx 64:%-10lu", MCF_FEC_RMON_R_P64);
  printf("      Rx 65-127:%-10lu", MCF_FEC_RMON_R_P65TO127);
  printf("     Rx 128-255:%-10lu\n", MCF_FEC_RMON_R_P128TO255);
  printf("     Rx 256-511:%-10lu", MCF_FEC_RMON_R_P256TO511);
  printf("    Rx 511-1023:%-10lu", MCF_FEC_RMON_R_512TO1023);
  printf("   Rx 1024-2047:%-10lu\n", MCF_FEC_RMON_R_1024TO2047);
  printf("      Rx >=2048:%-10lu", MCF_FEC_RMON_R_P_GTE2048);
  printf("      Rx Octets:%-10lu", MCF_FEC_RMON_R_OCTETS);
  printf("     Rx Dropped:%-10lu\n", MCF_FEC_IEEE_R_DROP);
  printf("    Rx frame OK:%-10lu", MCF_FEC_IEEE_R_FRAME_OK);
  printf("   Rx CRC error:%-10lu", MCF_FEC_IEEE_R_CRC);
  printf(" Rx Align error:%-10lu\n", MCF_FEC_IEEE_R_ALIGN);
  printf("  FIFO Overflow:%-10lu", MCF_FEC_IEEE_R_MACERR);
  printf("Rx Pause Frames:%-10lu", MCF_FEC_IEEE_R_FDXFC);
  printf("   Rx Octets OK:%-10lu\n", MCF_FEC_IEEE_R_OCTETS_OK);
  printf(" Tx Unaccounted:%-10lu", MCF_FEC_RMON_T_DROP);
  printf("Tx Packet Count:%-10lu", MCF_FEC_RMON_T_PACKETS);
  printf("   Tx Broadcast:%-10lu\n", MCF_FEC_RMON_T_BC_PKT);
  printf("   Tx Multicast:%-10lu", MCF_FEC_RMON_T_MC_PKT);
  printf("CRC/Align error:%-10lu", MCF_FEC_RMON_T_CRC_ALIGN);
  printf("   Tx Undersize:%-10lu\n", MCF_FEC_RMON_T_UNDERSIZE);
  printf("    Tx Oversize:%-10lu", MCF_FEC_RMON_T_OVERSIZE);
  printf("    Tx Fragment:%-10lu", MCF_FEC_RMON_T_FRAG);
  printf("      Tx Jabber:%-10lu\n", MCF_FEC_RMON_T_JAB);
  printf("  Tx Collisions:%-10lu", MCF_FEC_RMON_T_COL);
  printf("          Tx 64:%-10lu", MCF_FEC_RMON_T_P64);
  printf("      Tx 65-127:%-10lu\n", MCF_FEC_RMON_T_P65TO127);
  printf("     Tx 128-255:%-10lu", MCF_FEC_RMON_T_P128TO255);
  printf("     Tx 256-511:%-10lu", MCF_FEC_RMON_T_P256TO511);
  printf("    Tx 511-1023:%-10lu\n", MCF_FEC_RMON_T_P512TO1023);
  printf("   Tx 1024-2047:%-10lu", MCF_FEC_RMON_T_P1024TO2047);
  printf("      Tx >=2048:%-10lu", MCF_FEC_RMON_T_P_GTE2048);
  printf("      Tx Octets:%-10lu\n", MCF_FEC_RMON_T_OCTETS);
  printf("     Tx Dropped:%-10lu", MCF_FEC_IEEE_T_DROP);
  printf("    Tx Frame OK:%-10lu", MCF_FEC_IEEE_T_FRAME_OK);
  printf(" Tx 1 Collision:%-10lu\n", MCF_FEC_IEEE_T_1COL);
  printf("Tx >1 Collision:%-10lu", MCF_FEC_IEEE_T_MCOL);
  printf("    Tx Deferred:%-10lu", MCF_FEC_IEEE_T_DEF);
  printf(" Late Collision:%-10lu\n", MCF_FEC_IEEE_T_LCOL);
  printf(" Excessive Coll:%-10lu", MCF_FEC_IEEE_T_EXCOL);
  printf("  FIFO Underrun:%-10lu", MCF_FEC_IEEE_T_MACERR);
  printf("  Carrier Error:%-10lu\n", MCF_FEC_IEEE_T_CSERR);
  printf("   Tx SQE Error:%-10lu", MCF_FEC_IEEE_T_SQE);
  printf("Tx Pause Frames:%-10lu", MCF_FEC_IEEE_T_FDXFC);
  printf("   Tx Octets OK:%-10lu\n", MCF_FEC_IEEE_T_OCTETS_OK);
}

NBUF *Xbuf;

/* getXmitBuffer():
 * Return a pointer to the buffer that is to be used for transmission of
 * the next packet.  Since the monitor's driver is EXTREMELY basic,
 * there will only be one packet ever being transmitted.  No need to queue
 * up transmit packets.
 */
uchar *
getXmitBuffer(void)
{
	Xbuf = nbuf_tx_allocate();
	if(Xbuf)
	{
		return(Xbuf->data);
	}
	return 0;
}

/* sendBuffer():
 * Send out the packet assumed to be built in the buffer returned by the
 * previous call to getXmitBuffer() above.
 */
int
sendBuffer(int length)
{
#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_OUTGOING)
		printPkt((struct ether_header *)Xbuf->data,length,ETHER_OUTGOING);
#endif

	WATCHDOG_MACRO;

	/* Bump up the packet length to a minimum of 64 bytes.
	 */
	if (length < 64)
		length = 64;

	/* Add the code that will tickle the device into sending out the
	 * buffer that was previously returned by getXmitBuffer() above...
	 */
	Xbuf->length = length;

	/* Set Frame ready for transmit */
	Xbuf->status |= MCF_FEC_TxBD_R;

	/* Mark the buffer as not in use */
	nbuf_tx_release(Xbuf);

	/* Indicate that there has been a transmit buffer produced */
	MCF_FEC_TDAR = MCF_FEC_TDAR_X_DES_ACTIVE;

	/* The hardware will mark the buffer as empty once it's transmitted */

	EtherXFRAMECnt++;
	return(length);
}

/* DisableEtherdev():
 * Fine as it is...
 */
void
DisableEtherdev(void)
{
	enreset();
}

/* extGetIpAdd():
 * If there was some external mechanism (other than just using the
 * IPADD shell variable established in the monrc file) for retrieval of
 * the board's IP address, then do it here...
 */
char *
extGetIpAdd(void)
{
	return((char *)0);
}

/* extGetEtherAdd():
 * If there was some external mechanism (other than just using the
 * ETHERADD shell variable established in the monrc file) for retrieval of
 * the board's MAC address, then do it here...
 */
char *
extGetEtherAdd(void)
{
	return((char *)0);
}

/*
 * polletherdev():
 * Called continuously by the monitor (ethernet.c) to determine if there
 * is any incoming ethernet packets.
 *
 * NOTES:
 * 1. This function must be reentrant, because there are a few cases in
 *    processPACKET() where pollethernet() may be called.
 * 2. It should only process one packet per call.  This is important
 *    because if allowed to stay here to flush all available packets,
 *    it may starve the rest of the system (especially in cases of heavy
 *    network traffic).
 * 3. There are cases in the monitor's execution that may cause the
 *    polling polletherdev() to cease for several seconds.  Depending on
 *    network traffic, this may cause the input buffering mechanism on
 *    the ethernet device to overflow.  A robust polletherdev() function
 *    should support this gracefully (i.e. when the error is detected,
 *    attempt to pass all queued packets to processPACKET(), then do what
 *    is necessary to clear the error).
 */
int
polletherdev(void)
{
	int pktcnt = 0, pktlen = 0;
	NBUF *pNbuf;

	WATCHDOG_MACRO;

	while(nbuf_rx_next_ready()) {
		/* Get the next buffer received by the FEC */
		pNbuf = nbuf_rx_allocate();
		if (pNbuf == 0) {
			printf("rx_allocate1 failed\n");
			return(pktcnt);
		}
		if (!(pNbuf->status & MCF_FEC_RxBD_L))
		{
			do
			{	/* Free the current buffer */
				nbuf_rx_release(pNbuf);

				/* Get the next buffer of this frame */
				pNbuf = nbuf_rx_allocate();
				if (pNbuf == 0) {
					printf("rx_allocate2 failed\n");
					return(pktcnt);
				}
	
				/* Keep going until we find the last one */
			} while (!(pNbuf->status & MCF_FEC_RxBD_L));

			/* Free the last buffer of this frame */
			nbuf_rx_release(pNbuf);
			return(pktcnt);
		}
	
		/* Adjust the length by the ethernet header size */
		pktlen = pNbuf->length - (4 + 6 + 6 + 2);
	
		/* Call the appriopriate handler */
		processPACKET((struct ether_header *) pNbuf->data, pktlen);
	
		nbuf_rx_release(pNbuf);
		pktcnt++;

		MCF_FEC_RDAR = MCF_FEC_RDAR_R_DES_ACTIVE;
	}
	return(pktcnt);
}

#endif
