/* reset.s:
 *
 * First bit of boot code run by the processor.
 *
 */
	.file   "reset.s"

	.extern start
	.extern MonStack
	.extern stkinit
	.extern init_main
	.global reset
	.global warmstart
	.global coldstart
	.global warmstartcontinue
	.global resetSP
	.global resetPC

#include "warmstart.h"
#include "cpuio.h"
#include "config.h"

/* Leave some room at the top of stack for start up variables */
#define INITIAL_TOP_OF_STACK   (INT_SRAM_ADDRESS + INT_SRAM_SIZE - 32)
#define RESET_VECTOR_STORE     (INT_SRAM_ADDRESS + INT_SRAM_SIZE - 16)

/*********************************************************************
 *
 * coldstart:
 * The reset point.
 *
 * The flash chip select CS0 is active at address zero.  We don't
 * need to do anything funky here (like the csb360 port) as our
 * flash memory map is at 0.
 *
 */
reset:
resetSP:	/* Initial SP = end of internal SRAM */
	.long       INITIAL_TOP_OF_STACK

resetPC:	/* Initial PC */
	.long       coldstart

	/* Note: if we were running with a flash-based vector table, then
	 * it would reside here.  This version of the monitor builds its
	 * vector table into RAM (and sets VBR accordingly) when vinit()
	 * is called; hence, for a brief amount of time at bootup the
	 * cpu is running without the ability to deal with exceptions
	 * (other than RESET).
	 */

coldstart:
	/* - Invalidate and disable caches
	 * - Load INITIALIZE (defined in warmstart.h) into whereever it
	 *   needs to be put so that warmstart sees it as an argument
	 *   passed to it.
	 *
	 * - Jump to warmstart
	 */

	move.l      %d0,%d6
	move.l      %d1,%d7

	/* Warmstart puts the start type into d5, given we are not calling
	 * from c, we'll do the same, and jump to warmstartcontinue
	 * instead.
	 */

	move.l      #INITIALIZE,%d5
	jmp         warmstartcontinue

ram_coldstart:

	move.l      #0,%d6
	move.l      #0,%d7
	move.l      #INITIALIZE,%d5
	jmp         warmstartcontinue
	
/*********************************************************************
 *
 * moncomptr:
 * Pointer to the moncom function, used to link application to monitor.
 * Refer to umon_main/target/core/moncomptr.S
 */
#include "moncomptr.S"

/*********************************************************************
 *
 * etheraddr:
 * Location that could be used to store a fixed MAC address.
 * Refer to umon_main/target/core/etheraddr.S.
 * NOTE:
 * This should only be included in flash-resident code, then
 * the address in flash should be accessible to ram-based versions of
 * uMon via tags in the linker map file.
 */
#include "etheraddr.S"

/*********************************************************************
 *
 * alttfsdevtbl.S:
 * Location that could be used to store an "alternate" TFS device
 * table for use by the "tfs cfg" command.
 * Refer to umon_main/target/core/alttfsdevtbl.S.
 * NOTE:
 * This should only be included in flash-resident code, then
 * the address in flash should be accessible to ram-based versions of
 * uMon via tags in the linker map file.
 */

/*
 * Because the 5329 has 2 tfs devices initialized (flash and ram) we
 * need an extra dev table cfg block in flash.  See the porting notes,
 * which suggest that we don't include alttfsdevtbl.S as per the
 * template port, but rather copy the contents in and make things
 * as large as is required.
 */

	.globl	alt_tfsdevtbl_base

	.balign	0x10

alt_tfsdevtbl_base:
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff

	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff

/*********************************************************************
 *
 * warmstart:
 * A point callable by C, as warmstart(int type), where
 * 'type' is one of the values defined in warmstart.h.
 */
warmstart:

	/* Warmstart is callable by C, so poke the startup type into d5 which
	 * we won't play with until we retrieve it to pass to start
	 */
	move.l      %sp@(4),%d5
	move.l      #0,%d6
	move.l      #0,%d7

warmstartcontinue:
	move.w      #0x2700,%SR

	/* Minimal CPU/IO Initialization (i.e. chip selects) here.
	 * Stay in assembler here if possible.
	 */

	/* Disable watchdog
	 */
	move.l      #0xFC040016,%a0
	move.w      #0,(%a0)

	/* Invalidate and disable the cache...
	 */
	move.l      #0x01000000,%d0
	movec       %d0,%CACR

	move.l      #0x00000000,%d0
	movec       %d0,%ACR0
	move.l      #0x00000000,%d0
	movec       %d0,%ACR1

	/* Initialize RAMBAR1: locate SRAM and validate it - Allow FEC access
	 */
	move.l      #INT_SRAM_ADDRESS,%d0
	add.l       #0x221,%d0
	movec       %d0,#0xc05 /* RAMBAR1 */

	/* Save off intial D0 and D1 to top of startup stack
	 */ 
	move.l      #RESET_VECTOR_STORE,%a0
	move.l      %d5,(%a0)+
	move.l      %d6,(%a0)+
	move.l      %d7,(%a0)+

	/* Setup a temporary stack in core sram that lets us call 
	 * initialization routines in c.
	 */
	move.l      #INITIAL_TOP_OF_STACK,%sp

	/* Initialize the board to a startup state.
	 */
	jsr         init_main

/*********************************************************************
 *
 * SPInit:
 * Establish a stack frame.
 */
SPInit:
	/* First initialize stack to point to MonStack+MONSTACKSIZE+64
	 * (an address that is outside the MonStack array), and then
	 * call stkinit().  This loads the MonStack[] array with a known
	 * pattern that allows uMon to later analyze the running stack usage.
	 */

	move.l      #MonStack,%d0
	add.l       #MONSTACKSIZE, %d0
	add.l       #64, %d0
	move.l      %d0, %sp
	jsr stkinit

	/* Next, re-initialize stack to point to MonStack+MONSTACKSIZE...
	 * This is important because other portions of the code
	 * assume this is where the stack resides.
	 */
	move.l      #MonStack,%d0
	add.l       #MONSTACKSIZE, %d0
	move.l      %d0, %sp

/*********************************************************************
 *
 * gotoC:
 * This code jumps to the start() function in the monitor
 * and should never return.
 */
gotoC:
	/* - Retrieve the variable passed to warmstart and place it
	 *   whereever it needs to be so that the start() function in
	 *   'C' sees it as an argument.  Note that if FORCE_BSS_INIT
	 *   is defined, then start() will ignore this argument.
	 * - Branch to start().
	 */

	/* Transition to C...
	 */

	move.l      #RESET_VECTOR_STORE, %a0
	move.l      (%a0),-(%sp)
	jsr         start

    /* Should never get here: */
badnews:
	jmp coldstart

