/* etherdev.c:
 *
 * Ethernet driver for TEMAC.
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:	Ed Sutter
 * email:	esutter@lucent.com
 * phone:	908-582-2351
 */

#include "config.h"
#if INCLUDE_ETHERNET
#include "stddefs.h"
#include "ether.h"
#include "phy.h"
#include "xparameters.h"
#include "xlltemac.h"
#include "xllfifo.h"

extern int atoi(char *);
extern char *getenv(char *);
extern int printf(char * fmt, ...);

#define RXBUFTOT 4
#define EBUFSZ	1600

XLlTemac_Config		*TemacCfg;
XLlTemac			TemacInst;
XLlFifo				FifoInst;

char *enetadd, *ipadd;

int rxidx, phyAddr;
char tx_buf[EBUFSZ];
char rx_buf[RXBUFTOT][EBUFSZ];

int	EtherFifoError;
int	EtherRxRejectError;
int EtherRxFifoOvrError;

/*
 * PHY startup configuration information:
 * Each of the PHYs within the QUAD PHY can have its own startup (i.e.
 * at reset) configuration.  
 *
 * We are initially working with a direct connection from TEMAC to
 * the PHY at PHYADDR 4. This is PHY0 of the second QUAD PHY.
 * So, referring to pg 26 of schematic (to get the bit settings)
 * and table 37 of the MARVELL 88E1145 datasheet (to understand what
 * the bit settings are for)...
 *
 * 	PHYADDR:			00100		(PhyAddr: 4)
 * 	HWCFG_MODE:			1111		(GMII/MII to Copper)
 * 	ANEG:				1110		(aneg on, advertise all, prefer Master)
 *  CONFIG3:			0111
 * 	   PHYADDR[4]:		0
 * 	   ENA_XC:			1			(auto crossover is enabled)
 * 	   DIS_FC:			1
 * 	   DIS_SLEEP:		1
 *  CONFIG4:			xx00		(x=reserved)
 *     SEL_TWSI:		0			(don't use TWSI, use MDC/MDIO)
 *     ENA_PAUSE:		0
 * 	GCONFIG0:			1011
 *    DIS_DTE:			1
 *    75/50Ohm:			0
 *    1/4 MDIO::		1
 *    DIS_125CLK:		1
 * 	GCONFIG1:			0111
 *    LED_TX Blink:		0
 *    PowerDown:		1
 *    SIG_DET:			1
 *    INT_POL:			1
 */
/* phySpeed():
 * Return 0 if link is not up or speed/duplex is not resolved;
 * else return 1 and populate the speed and fdup values appropriately.
 */
int
phySpeed(int *speed, int *fdup)
{
	ushort reg;

	/* Read the PHY Specific Status Register.  If link is up and 
	 * speed/duplex are resolved, then populate speed and fdup and
	 * return 1; else return 0.
	 */
	phy_read(phyAddr,17,&reg);
	if ((reg & 0x0c00) != 0x0c00)
		return(0);

	if (speed) {
		switch(reg & 0xc000) {
			case 0x8000:
				*speed = 1000;
				break;
			case 0x4000:
				*speed = 100;
				break;
			case 0x0000:
				*speed = 10;
				break;
			default:
				*speed = 0;
				break;
		}
	}
	if (fdup)
		*fdup = ((reg & 0x2000) == 0x2000);
	return(1);
}


void
enreset(void)
{
}

void
enstop(void)
{
    /*
     * Stop and reset both the fifo and the temac the devices
     */
    XLlTemac_Stop(&TemacInst);
    XLlFifo_Reset(&FifoInst);
    XLlTemac_Reset(&TemacInst, XTE_NORESET_HARD);
}

int
eninit(void)
{
	u32 Rdy;
	char *env;
	volatile int i;
	int Status, speed, fdup, try;
	XLlTemac_Config *MacCfgPtr;
	u16 reg, TemacDeviceId = XPAR_LLTEMAC_0_DEVICE_ID;

	env = getenv("PHYADDR");
	if (env)
		phyAddr = atoi(env);
	else
		phyAddr = 4;

	printf("Enet startup: using phyaddr %d\n",phyAddr);

	/*************************************/
	/* Setup device for first-time usage */
	/*************************************/

	// Initialize the FIFO and TEMAC instance
	MacCfgPtr = XLlTemac_LookupConfig(TemacDeviceId);
	if (MacCfgPtr == NULL) {
		printf("XLlTemac_LoopConfig(%d) failed\n",TemacDeviceId);
		return(XST_FAILURE);
	}

	Status = XLlTemac_CfgInitialize(&TemacInst,MacCfgPtr,
		MacCfgPtr->BaseAddress);
	if (Status != XST_SUCCESS) {
		printf("Error in initialize\n");
		return XST_FAILURE;
	}

	// Make sure we're in FIFO mode...
	if (!XLlTemac_IsFifo(&TemacInst)) {
		printf("Device HW not configured for FIFO direct mode\n");
		return XST_FAILURE;
	}

	XLlFifo_Initialize(&FifoInst,
		XLlTemac_LlDevBaseAddress(&TemacInst));

	/*****************************************************************/
	// Following the documentation in the xps_ll_fifo (v1.00a) document
	// (pg 23)...
	reg = XLlFifo_Status(&FifoInst);
	XLlFifo_IntClear(&FifoInst,0xffffffff);
	reg = XLlFifo_Status(&FifoInst);
	if (reg != 0) 
		printf("WARNING: FIFO ISR did not clear\n");
	reg = XLlFifo_ReadReg(FifoInst.BaseAddress, XLLF_IER_OFFSET);
	if (reg != 0)
		printf("WARNING: FIFO IER not zero\n");
    reg = XLlFifo_TxVacancy(&FifoInst);
	if (reg != 0x1fe)
		printf("WARNING: FIFO xmit vacancy != 0x1fe (0x%lx)\n",reg);
    reg = XLlFifo_RxOccupancy(&FifoInst);
	if (reg != 0)
		printf("WARNING: FIFO rcv occupancy != 0 (0x%lx)\n",reg);
	//
	// Done that which is specified in table 16 of XPS_LL_FIFO document.
	/*****************************************************************/


	// Set the MAC  address
	Status = XLlTemac_SetMacAddress(&TemacInst,(u8 *)BinEnetAddr);
	if (Status != XST_SUCCESS) {
		printf("Error setting MAC address\n");
		return XST_FAILURE;
	}

	// Make sure the hard temac is ready.
	Rdy = XLlTemac_ReadReg(TemacInst.Config.BaseAddress,XTE_RDY_OFFSET);

	try = 100000;
	while ((Rdy & XTE_RDY_HARD_ACS_RDY_MASK) == 0) {
		Rdy = XLlTemac_ReadReg(TemacInst.Config.BaseAddress,
				       XTE_RDY_OFFSET);
		if (--try == 0) {
			printf("Temac ready-poll timeout\n");
			return(-1);
		}
	}

	/* Wait for speed and duplex mode to be resolved...
	 */
	for(try=100000;try>0;try--) {
		if (phySpeed(&speed,&fdup))
			break;
	}
	if (try == 0) {
		printf("Ethernet: Link down\n");
		return(-1);
	}

	printf("Enet speed: %dBaseT, %s-duplex\n",speed,fdup?"full":"half");
	XLlTemac_SetOperatingSpeed(&TemacInst, speed);

	/*
	 * Setting the operating speed of the MAC needs a delay.  There
	 * doesn't seem to be register to poll, so please consider this
	 * during your application design.
	 */
	for(i=0;i<2000000;i++);

	/* Enable reception of multicast...
	 */
	XLlTemac_SetOptions(&TemacInst, XTE_MULTICAST_OPTION);

    /*
     * Start the TEMAC device
     */
    XLlTemac_Start(&TemacInst);

	rxidx = 0;

	return(0);
}


int
EtherdevStartup(int verbose)
{
	enreset();
	EtherFifoError = 0;
	EtherRxRejectError = 0;
	EtherRxFifoOvrError = 0;
	return(eninit());
}

/* getXmitBuffer():
 * Return a pointer to the buffer that is to be used for transmission of
 * the next packet.  Since the monitor's driver is EXTREMELY basic,
 * there will only be one packet ever being transmitted.  No need to queue
 * up transmit packets.
 */
uchar *
getXmitBuffer(void)
{
	return((uchar *)tx_buf);
}

/* sendBuffer():
 * Send out the packet assumed to be built in the buffer returned by the
 * previous call to getXmitBuffer() above.
 */
int
sendBuffer(int length)
{
	u32 fStat;
	int FifoFreeBytes, Attempts, freelen;
	XStatus Status = XST_NO_DATA;

#if INCLUDE_ETHERVERBOSE
    if (EtherVerbose &  SHOW_OUTGOING)
        printPkt((struct ether_header *)tx_buf,length,ETHER_OUTGOING);
#endif

	if (EtherXFRAMECnt > 0) {
		// We're here because EtherXFRAMECnt is greater than zero.
		// This means that a packet has been sent, so we must
		// make sure that the previous packet send has completed
		// before starting up another packet send...
		Attempts = 100000;
		do {
			if (--Attempts <= 0) {
				printf("sendBuffer: TxComplete timeout\n");
				return(-1);
			}
			fStat = XLlFifo_Status(&FifoInst);
	
			if (fStat & XLLF_INT_TC_MASK) {
				XLlFifo_IntClear(&FifoInst,XLLF_INT_TC_MASK);
				Status = XST_SUCCESS;
				break;
   			}
			if (fStat & XLLF_INT_ERROR_MASK) {
				printf("fStat: 0x%08lx\n",fStat);
				Status = XST_FIFO_ERROR;
				if (fStat & XLLF_INT_RPURE_MASK) {
					enstop();
					eninit();
				}
			}
		} while (Status == XST_NO_DATA);
		if (Status != XST_SUCCESS) {
			printf("sendBuffer failed: %d\n",Status);
		}
	}

    // Wait for enough room in FIFO to become available
	Attempts = 100000;
	freelen = length/4 + 1;	// XLlFifo_TxVacancy() returns # of 32-bit vals.
    do {
        FifoFreeBytes = XLlFifo_TxVacancy(&FifoInst);
        if (--Attempts <= 0) {
			printf("sendBuffer: TxVacancy timeout\n");
            return(-1);  
		}
    } while (FifoFreeBytes < freelen);

    // Write the frame data to FIFO
    XLlFifo_Write(&FifoInst, tx_buf, length);

    // Initiate transmit
    XLlFifo_TxSetLen(&FifoInst, length);

	// Note:
	// We don't wait for completion of the transmission here, we do
	// it at the top of the next call to sendBuffer().
	EtherXFRAMECnt++;
	return(0);
}

/* DisableEtherdev():
 */
void
DisableEtherdev(void)
{
    /*
     * Stop device
     */
	XLlTemac_Stop(&TemacInst);
}


/* polletherdev():
 * Called continuously by the monitor (ethernet.c) to determine
 * if there is any incoming ethernet packets.
 */

int
polletherdev(void)
{
	int len, giveup;
	u32 estat, fstat;
    struct ether_header *ehp;

	fstat = XLlFifo_Status(&FifoInst);
	estat = XLlTemac_Status(&TemacInst);

#if 0
	if (estat & XTE_INT_ALL_MASK) {
		printf("POLL_XTE_INT: 0x%lx\n",estat & XTE_INT_ALL_MASK);
		XLlTemac_IntClear(&TemacInst,estat);
	}
#endif

	if (estat & XTE_INT_RXRJECT_MASK) {
		EtherRxRejectError++;
		printf("EnetPoll: RxReject err\n");
		XLlTemac_IntClear(&TemacInst,XTE_INT_RXRJECT_MASK);
	}
	if (estat & XTE_INT_RXFIFOOVR_MASK) {
		EtherRxFifoOvrError++;
		printf("EnetPoll: RxfifoOvr err\n");
		XLlTemac_IntClear(&TemacInst,XTE_INT_RXFIFOOVR_MASK);
	}


	if ((fstat & XLLF_INT_RC_MASK) == 0)
		return(0);

	giveup = 10000;
	while(estat == 0) {
		estat = XLlTemac_Status(&TemacInst);
		if (--giveup == 0) {
			printf("EnetPoll: HERE\n");
			return(0);
		}
	}

	XLlTemac_IntClear(&TemacInst,XTE_INT_RC_MASK);
	if ((estat & XTE_INT_RC_MASK) == 0) {
		printf("EnetPoll: HERE_1\n");
		return(0);
	}

	XLlFifo_IntClear(&FifoInst,XLLF_INT_RC_MASK);
	XLlFifo_Status(&FifoInst);	// xps_ll_fifo.pdf says to read it again
								// (after the clear)

	if (fstat & XLLF_INT_ERROR_MASK) {
		EtherFifoError++;
		printf("POLL: xllf interr 0x%08lx\n",fstat);
	}

	/* Poll the "Receive Data FIFO Occupancy" register to check for
	 * an incoming packet.  This register value is only updated after
	 * packet is successfully received and thereefore can be used to
	 * determine if a receive packet is ready to be processed
	 * (text taken from xps_ll_fifo (v1.00a) document).
	 */
    if (XLlFifo_RxOccupancy(&FifoInst) == 0)
		return(0);

   	len = XLlFifo_RxGetLen(&FifoInst);

	if ((len > EBUFSZ) || (len <= 0)) {
		printf("enet rcvlen oor: 0x%x\n",len);
		return(0);
	}

   	XLlFifo_Read(&FifoInst, rx_buf[rxidx], len);

    ehp = (struct ether_header *)rx_buf[rxidx];
	if (++rxidx == RXBUFTOT)
		rxidx = 0;

	processPACKET(ehp,len);

	return(1);
}

void
ShowEtherdevStats(void)
{
	u32 reg;
	int speed, fdup;

	printf("Link is ");
	if (phySpeed(&speed,&fdup) == 0) {
		printf("down\n");
	}
	else {
		printf("up, %d MBps, %s-duplex\n",speed,fdup?"full":"half");
		printf("Promiscuous mode: ");
		reg = XLlTemac_ReadIndirectReg(TemacInst.Config.BaseAddress,
				XTE_AFM_OFFSET);
		if (reg & XTE_AFM_PM_MASK)
			printf("on\n");
		else
			printf("off\n");

		printf("Multicast rejection: ");
		reg = XLlTemac_ReadReg(TemacInst.Config.BaseAddress,
				XTE_RAF_OFFSET);
		if (reg & XTE_RAF_MCSTREJ_MASK)
			printf("on\n");
		else
			printf("off\n");
		
	}
		
#if 0
	{
	u32 rcw0, rcw1, fcc;
	u32 emmc, phyc, mc;
	rcw0 = XLlTemac_ReadIndirectReg(TemacInst.Config.BaseAddress,
		XTE_RCW0_OFFSET);
	rcw1 = XLlTemac_ReadIndirectReg(TemacInst.Config.BaseAddress,
		XTE_RCW1_OFFSET);
	fcc = XLlTemac_ReadIndirectReg(TemacInst.Config.BaseAddress,
		XTE_FCC_OFFSET);
	emmc = XLlTemac_ReadIndirectReg(TemacInst.Config.BaseAddress,
		XTE_EMMC_OFFSET);
	phyc = XLlTemac_ReadIndirectReg(TemacInst.Config.BaseAddress,
		XTE_PHYC_OFFSET);
	mc = XLlTemac_ReadIndirectReg(TemacInst.Config.BaseAddress,
		XTE_MC_OFFSET);

	printf("RCW0: 0x%08lx, RCW1: 0x%08lx, FCC: 0x%08lx\n",rcw0,rcw1,fcc);
	printf("EMMC: 0x%08lx, PHYC: 0x%08lx, MC:  0x%08lx\n",emmc,phyc,mc);
	}
#endif

}

/* disablePromiscuousReception():
 * Provide the code that disables promiscuous reception.
 */
void
disablePromiscuousReception(void)
{
	XStatus stat;

	printf("disablePromiscuouseReception()\n");

    XLlTemac_Stop(&TemacInst);
	stat = XLlTemac_ClearOptions(&TemacInst,XTE_PROMISC_OPTION);
	if (stat != XST_SUCCESS) {
		printf("stat: %d\n",(int)stat);
		return;
	}
    XLlTemac_Start(&TemacInst);
}

/* enablePromiscuousReception():
 * Provide the code that enables promiscuous reception.
 */
void
enablePromiscuousReception(void)
{
	XStatus stat;

	printf("enablePromiscuouseReception()\n");

    XLlTemac_Stop(&TemacInst);
	stat = XLlTemac_SetOptions(&TemacInst,XTE_PROMISC_OPTION);
	if (stat != XST_SUCCESS) {
		printf("stat: %d\n",(int)stat);
		return;
	}
    XLlTemac_Start(&TemacInst);
}

/* disableBroadcastReception():
 * Provide the code that disables broadcast reception.
 */
void
disableBroadcastReception(void)
{
	XStatus stat;

//	printf("disableBroadcastReception()\n");

	stat = XLlTemac_ClearOptions(&TemacInst,XTE_BROADCAST_OPTION);
	if (stat != XST_SUCCESS) {
//		printf("stat: %d\n",(int)stat);
	}
}

/* enableBroadcastReception():
 * Provide the code that enables broadcast reception.
 */
void
enableBroadcastReception(void)
{
	XStatus stat;

//	printf("enableBroadcastReception()\n");

	stat = XLlTemac_SetOptions(&TemacInst,XTE_BROADCAST_OPTION);
	if (stat != XST_SUCCESS) {
//		printf("stat: %d\n",(int)stat);
	}
}

char *
extGetIpAdd(void)
{
    return((char *)0);
}

char *
extGetEtherAdd(void)
{
    return((char *)0);
}

int
enselftest(int verbose)
{
	return(1);
}


#if INCLUDE_PHY

int
phy_read(int phynum, int reg, unsigned short *value)
{
	//printf("phyr(reg=%04x)\n",reg);
	XLlTemac_PhyRead(&TemacInst,(Xuint32)phynum,(Xuint32)reg,(Xuint16 *)value);
	return(0);
}

int
phy_write(int phynum, int reg, unsigned short value)
{
	//printf("phyw(reg=%04x,val=%04x)\n",reg,value);
	XLlTemac_PhyWrite(&TemacInst,(Xuint32)phynum,(Xuint32)reg,(Xuint16)value);
	return(0);
}

#define PHYWRITE	1
#define PHYREAD		2
#define PHYVERBOSE	3
#define PHYQUIET	4

struct phyr_info {
	int phynum;
	int reg;
	unsigned short *value;
};

struct phyw_info {
	int phynum;
	int reg;
	unsigned short value;
};

int
portCmd(int cmd, void *arg)
{
	int rc = -1;
	struct phyr_info *phyr;
	struct phyw_info *phyw;
	static int verbose = 0;

	if (cmd == PHYREAD) {
		phyr = (struct phyr_info *)arg;
		rc = phy_read(phyr->phynum,phyr->reg,phyr->value);
		if (verbose)
			printf("PHYREAD(%d,%d) = 0x%04x (rc=%d)\n",
				phyr->phynum,phyr->reg,*phyr->value,rc);
	}
	else if (cmd == PHYWRITE) {
		phyw = (struct phyw_info *)arg;
		rc = phy_write(phyw->phynum,phyw->reg,phyw->value);
		if (verbose)
			printf("PHYWRITE(%d,%d,0x%04x) (rc=%d)\n",
				phyw->phynum,phyw->reg,phyw->value,rc);
	}
	else if (cmd == PHYVERBOSE)
		verbose = 1;
	else if (cmd == PHYQUIET)
		verbose = 0;
	else
		printf("portCmd(), invalid cmd: %d\n",cmd);

	return(rc);
}

#endif
#endif
